WARNING
COmpare  replace metadata code with that found in the regressed version
#################################################################
##
## Kirk has a different way to treat the filtering of the data.
## 1) For the Raw data  compute the TOTAL RAW mean counts as
## total_c sum(READS_C,G)
## avg_reads = SUM(total_c)/N
## 
## 2) For the Imputed data from Filer
## total_imp_c = SUM(READS_c,g)
## New_impute_cg = ( 10.0 * avg_reads * READS_c,g ) / total_imp_c
##
## See book causal networks. pg 27
##
## 3) Remove genes with fewer than 5908 cells with scaled reads < 1
## 4) Remove those cells that were not tumor cells

##
## BUILD a new datafrae as follows. Columns CELLS MOUSE SEX DATE VEHICLE gene1 gene2 ....genep
## The gene values should be PRESCALED but not whitened. We can do thatlater
## The MOUSE value can be constructed form the CELL name

# Orig MetaData
##"Sex"   "Treatment"     "TenPC.hvg.ident"       "Node"

# MetaDataWithDate
##"nGene" "nUMI"  "Replicate"     "Treatment"     "Sex"   "Date"  "ID"

## So we have initially 16,600 Genes and 29,538 SingleCells
##
##################################################################

import time as tm
import pandas as pd
import numpy as np
import collections
import matplotlib.pyplot as plt
import math 
import scipy 
from scipy import stats
import sys
import sklearn.preprocessing

import sys

from numpy.random import seed
from numpy.random import randn
from statsmodels.graphics.gofplots import qqplot
import pylab
import scipy.stats as stats
from scipy import linalg


#from __future__ import print_function

#########################################################################

# Change these to suit
# Not all these files are always used

# Input range of cell types classified as TUMORS to INCLUDE
if len(sys.argv) >2:
    lowValue = int(sys.argv[1])
    highValue = int(sys.argv[2])
elif (len(sys.argv) == 2):
    lowValue = int(sys.argv[1])
    highValue = int(sys.argv[1])
else:
    lowValue = 1
    highValue = 10

print("Range of cells to keep is "+str(lowValue)+' to '+str(highValue))
if (highValue > 10):
    print("High Value may be too high should be 10 or less "+str(highValue))

if (lowValue < 1):
    print("Low Value may be too low should be 1 or greater "+str(lowValue))

rootdir = '/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/'
rawfilename = rootdir+'rawResults/RawDataMatrix.txt'
metafilenameDate = rootdir+'rawResults/MetaDataMatrixWithDate.txt'

filename = rootdir+'dcaResults/mean.tsv'

t0=tm.time()
rawdata = pd.read_csv(rawfilename,delim_whitespace=True)
data = pd.read_csv(filename,delim_whitespace=True)
print(tm.time()-t0)

##
## Some basic plots of Dayne's imputed data
##

datameans = pd.DataFrame(data.mean(axis=1))
datameans.reset_index(inplace=True)
datameans.columns=['GENE','MEANS']
#datameans.plot(kind='line',x='GENE',y='MEANS',label='Kirk: Imputed data: Column MEANS of imputed scRNAseq data')
#$datameans.xlabel=('Gene')
#plt.ylabel=('Mean')
#plt.savefig('imputedData_GeneMeanCounts.pdf')
#plt.show()

##
## How about the cell mean reads?
##

datacellmeans = pd.DataFrame(data.mean())
datacellmeans.reset_index(inplace=True)
datacellmeans.columns=['CELL','CELLMEANS']
datacellmeans.sort_values(by=['CELL'],inplace=True)
#datacellmeans.plot(kind='line',x='CELL',y='CELLMEANS',label='Column MEANS of imputed scRNAseq data')
#datacellmeans.xlabel=('Cell')
#datacellmeans.ylabel=('CellMean')
#plt.savefig('rawData_CellMeanCounts.pdf')
#plt.show()

######################################################################
## Compute gene MAD() values then keep 5000 top

datamad = pd.DataFrame(data.mad(axis=1))
datamad.reset_index(inplace=True)
datamad.columns=['GENE','MAD']
#datamad.plot(kind='line',x='GENE',y='MAD',label='Kirk: Imputed data: MAD() of imputed scRNAseq data')
#plt.ylabel=('Mean')
#plt.savefig('imputedData_GeneMeanCounts.pdf')
#plt.show()

## Sort data set  and include only the4 Top 5000 MAD values
datamad.sort_values(by='MAD',ascending=False,axis=0,inplace=True)
topGenes = datamad.head(5000)
topGenesList = topGenes['GENE'].tolist()
data.loc[topGenesList]

######################################################################
## Get list of TUMOR names for future use and proportions
## Filter-keep only Vehicle cells

def tumor_within_cellName(tumorname,cellname):
    return tumorname in cellname

def computeSizeFactor_cellname(cellname, counterObject):
    name = cellname.split('_')[0]
    numbcells = float(counterObject[name])
    total = float(sum(counterObject.values()))
    return numbcells/total

fulllistNames = list(data.loc[topGenesList].columns.values)
unique_tumors = set([ x.split('_')[0] for x in fulllistNames ])
#{'MB10', 'MB13', 'MB05', 'MB14', 'MB06', 'MB04', 'MB15', 'MB07', 'MB12', 'MB09'}
allentries_tumors_temp = list([ x.split('_')[0] for x in fulllistNames ])
from collections import Counter
counter = collections.Counter(allentries_tumors_temp)

## 1) Get some base data: Compute total ave read count from the Raw data
## just for comparisons
## We assume missingness is not to0 bad. But it seems we should QC the raw data 
## to get the following values  (?)

total_reads = float(rawdata.values.sum())
num_reads = float(rawdata.values.size)
ave_total_reads = total_reads / num_reads

##
## 2) Scale the reads of the imputed data set
## 

total_impute_c = pd.DataFrame(data.loc[topGenesList].mean(axis=0))
total_impute_c.columns=['TOTALC']

# Compute a Series: New_Impute_cg = impute_cg * (10*ave_total_reads) / total_imput_c
#cellmeans = pd.DataFrame(df_R.mean(axis=1) These all sum(genes) to a value of about 21019 per cell

total_correction_factor = 10.0 * ave_total_reads / total_impute_c['TOTALC']
data_new = data.loc[topGenesList].mul(total_correction_factor,axis=1)

# Confirm total reads per cell are the same
cellsums = data_new.sum(axis=0) # Values of 21019

# What about total reads per gene
genesums = data_new.sum(axis=1)
#genesums.hist(bins=20,range=[0,100000])
#plt.savefig('imputedData_GeneSums_postScale.pdf')
#plt.show()

###############################################################################3
## We want to remove genes that have fewer than 100 cells that are non-zero
## But what constitutes non-zero is somewhat arbitrary. We assume in the unscaled imputed 
## data that approx <= 10-3 is zero. However kirk think we should choose >= 1 as the scaled criteria 

# Keep only those that are nominal zeros (<1)
# Then do the counting. Any with more than 29538 - 100 = 29438 gets eliminated

x = data_new[ data_new < 1 ]
counts_thresh = x.count(axis=1).to_frame()     

cells_20percent =  data.shape[1] - int(0.20 *data.shape[1])

removeBoolean = counts_thresh[ counts_thresh > cells_20percent ] # approx 20%
removeBoolean.dropna(inplace=True)
listToRemove = removeBoolean.index.values
data_new.drop(listToRemove,inplace=True)

## Post QC Plot information
# Per gene
datameans = pd.DataFrame(data_new.mean(axis=1))
datameans.reset_index(inplace=True)
datameans.columns=['GENE','VARIANCE']
#datameans.plot(kind='line',x='GENE',y='VARIANCE',label='Imputed data: Column Scaled-MEANS of imputed scRNAseq data')
#datameans.xlabel=('Gene')
#datameans.ylabel=('Scaled-MEANS')
#plt.savefig('imputedData_GeneMeanCounts_removeMissingness.pdf')
#plt.show()

# Per Cell
datacellmeans = pd.DataFrame(data_new.mean())
datacellmeans.reset_index(inplace=True)
datacellmeans.columns=['CELL','CELLMEANS']
datacellmeans.sort_values(by=['CELL'],inplace=True)
#datacellmeans.plot(kind='line',x='CELL',y='CELLMEANS',label='Column MEANS of imputed scRNAseq data')
#datacellmeans.xlabel=('Cell')
#datacellmeans.ylabel=('CellMean')
#plt.savefig('rawData_CellMeanCounts_removeMissingness.pdf')
#plt.show()

## Gene must be > 0 for more than 100 cell2. Where zero is defined by
## scaledzero = 10-3 * (10*ave_total_reads) / total_imput_c

############################################################################################
# cell reads for al lmice have now been QC'd and scaled. We start filtering away non-vehicle, non-tumor data
# Then build data into the dataframe to apture some batch effects such as Date and Sex

## First Remove all CELLs that are not actually part of a Tumor using the MAP from Kirk
## /proj/kirklab/users/KirkFiles/
## We also only want VEHICLE cells

# NODENAMES THAT ARE NOT TUMORS: Remaining terms have the nomenclaturee NODE A, NODE B, etc
notTumors =['VascularFibroblasts','Astrocytes','Endothelial','Microglia','Neurons','Oligodendrocytes','Vascular']

metadata = pd.read_csv(metafilenameDate,delim_whitespace=True)

# Get list of cells that are NOT tumors
listNotTumorCells_df = metadata.ix[((metadata['ID'].isin(notTumors)))]
listNotTumorCells = list(listNotTumorCells_df.index.values)

## Now get list of cells from metadatawithdate that are NOT Vehicle for exclusion
listNotVehicleCells_df = metadata.ix[ metadata['Treatment'] != 'Vehicle']
listNotVehicleCdells = list(listNotVehicleCells_df.index.values)

# Union unique of exclude lists
excludeList = set(listNotTumorCells+listNotVehicleCdells)
data_new_tumor = data_new.copy()
data_new_tumor.drop(excludeList,axis=1,inplace=True)

cellNames = data_new_tumor.iloc[1].to_frame()
cellNamesList = cellNames.index.tolist()
geneNames = data_new_tumor.iloc[:,1].to_frame()
geneNamesList = geneNames.index.tolist()
geneNamesList_cell = geneNamesList

# get data from numpy and convert to df

data_new_tumor_scaled_T = pd.DataFrame(sklearn.preprocessing.scale(data_new_tumor.T, with_mean=True, with_std=True, copy=True).tolist())
data_new_tumor_scaled_T['CELL'] = cellNamesList
data_new_tumor_scaled_T.set_index('CELL',inplace=True)
data_new_tumor_scaled_T.columns = geneNamesList_cell

# Add metadata to the scaled data object (SEX,DATE,MOUSE)

scaledData = data_new_tumor_scaled_T.copy()
newcolumn = metadatadate.loc[scaledData.index.values][['Sex','Date','ID']]
##newcolumn.ID = newcolumn.ID.str.split(' ').str.get(1)
newcolumn.ID = [x.replace(' ','') for x in list(newcolumn.ID)]

# Now let's build a column that only has a mouse ID. We can convert to multiple rows of 
# indicator variables at some later time

mouseIndex = [ int(x.split('_')[0][2:]) for x in newcolumn.index ]
newcolumn['MOUSE']=mouseIndex

# Build final dataframe for disk storage
combinedScaledData = pd.merge(newcolumn,scaledData,left_index=True,right_index=True)
combinedScaledData.rename(columns={'Sex':'SEX','Date':'DATE','ID':'NODEID'},inplace=True)

outfilename = 'mouse_data_vehicle_tumor_scaled.tsv'
combinedScaledData.to_csv(outfilename,header=True,sep=" ",index=True)
################################################################
# might as well build the whitened data set too

X = data_new_tumor_scaled_T.values
Carray = data_new_tumor_scaled_T.cov().values
D,V = linalg.eigh(Carray)
indices = D.argsort()[::-1]
D,V = D[indices], V[:,indices]
Dhalf = np.sqrt(D)
DhalfInv = 1.0/Dhalf
DmatrixHalfInv = np.diag(DhalfInv)
CarrayInvHalf = V @ DmatrixHalfInv @ V.T
whiteX3 =  X @ CarrayInvHalf
newCovar = np.cov(whiteX3,rowvar=False)

##
## Must create a new DataFrame by hand
##
newDF = pd.DataFrame(whiteX3)
newDF.columns = geneNamesList
newDF['CELL']=data_new_tumor_scaled_T.index.values
newDF.set_index('CELL',inplace=True)

##
## Now add the metadata again
##
# Add metadata to the scaled data object (SEX,DATE,MOUSE)

newcolumn = metadatadate.loc[newDF.index.values][['Sex','Date','ID']]
# Fix the seemingly stupid choice for Node ID values
##newcolumn.ID = newcolumn.ID.str.split(' ').str.get(1)
newcolumn.ID = [x.replace(' ','') for x in list(newcolumn.ID)]

# Now let's build a column that only has a mouse ID. We can convert to multiple rows of 
# indicator variables at some later time

mouseIndex = [ int(x.split('_')[0][2:]) for x in newcolumn.index ]
newcolumn['MOUSE']=mouseIndex

# Build final dataframe for disk storage
combinedWhiteData = pd.merge(newcolumn,newDF,left_index=True,right_index=True)
combinedWhiteData.rename(columns={'Sex':'SEX','Date':'DATE','ID':'NODEID'},inplace=True)

outfilename = 'mouse_data_vehicle_tumor_scaled_whitened.tsv'
combinedWhiteData.to_csv(outfilename,header=True,sep=" ",index=True)




# READ IN THE XTILDE data # Classify the CELLS based on nearest neighbors algorithm.
# On input we expect a matrix of dim (nxp)
# n is num cells, p is num genes. We anticipate the input
# matrix (X) has been processed for group effects standardized to N(0,1)
#
# We cluster Xtilde but fit regression based on the current values of R
#
# NOTE: distances/indices are relative to the particular rows. BUT, since the row is in the training
# set it gets returned as a nearest neighbor. SO we want to set k = k+1 and ten exclude the 
# redundant terms

# generateNeighborMatrixClass.py

# takes about 15,000 secs to proces the clustering
# Maybe we should use a smaller training set to do the predictions
# May want to explore training on a subset of cells to minimize time

# Example input variables
#k=10
#inputfilename='100_MOUSEDATESEXTREATMENT_Xtilde_ZscoreOrder.tsv'
#outputIndices = 'dfIndices.tsv'
#outputDistances = 'dfDistances.tsv'

import pandas as pd
import numpy as np
#import math
#import matplotlib.pyplot as plt
#import sklearn.preprocessing
#from sklearn import preprocessing
#from sklearn.preprocessing import StandardScaler
from sklearn.neighbors import NearestNeighbors
import sys
import time as tm
from sklearn.model_selection import train_test_split
from sklearn.externals.joblib import Parallel, delayed

########################################################
def constructHeaderNames(kp1):
    headers=list()
    headers.append('CELL')
    #headers.append('CELLTARGET') # NOT TRUE based on NN bahavior
    for i in range(0,kp1):
       headers.append('NN'+str(i))
    return(headers)

########################################################

class generateNeighborMatrixClass(object):
    """Compute the K matrix. K is an (n*k) matrix (cells by neighbors
    We want indices and distance matrices. These matrices will be passed to
    a regression code for imputation
    Num neighbors: k+1: Because we also include the TARGET cell in the fit of k neighbors 
    output index file: 'dfIndices.tsv' 
    output distance file 'dfDistances.tsv'

    Attributes
    k = Number of nearest neighbors to fit
    infilename = filename for the Xtilde file returned from projectGroupDataClass
    outputIndeciesFilename = output indices filename
    outputDistancesFilename = output distances filename
    outputIndices = outputs object (not used by the pipeline) 
    outputDistances = outputs object (not used by the pipeline) 
    njobs = number of local threads to launch (16 or less is a good choice)
    """

    def __init__(self, k, infilename, outindex, outdistance, njobs=None):
        self.k = k
        self.inputfilename = infilename
        self.outputIndicesFilename = outindex
        self.outputDistancesFilename = outdistance
        self.outputIndices = pd.DataFrame()
        self.outputDistances = pd.DataFrame()
        self.njobs = njobs

    def reportIndices(self):
        return (self.outputIndices)
    def reportDistances(self):
        return (self.outputDistances)

    def reportParameters(self):
        print('input k neighbors '+str(self.k))
        print('input Xtilde '+self.inputfilename)
        print('output indexing '+self.outputIndicesFilename)
        print('output distances '+self.outputDistancesFilename)

    def writeClusterInformation(self):
        self.outputIndices.to_csv(self.outputIndicesFilename,sep=' ')
        self.outputDistances.to_csv(self.outputDistancesFilename,sep=' ')
        print('Writing of index/distance data is completed')

    def getIndexFilename(self):
        return self.outputIndicesFilename

    def getDistanceFilename(self):
       return self.outputDistancesFilename

    def runProcess(self):
        k=self.k
        inputfilename = self.inputfilename
        
        df_X = pd.read_table(inputfilename,delim_whitespace=True, index_col=0,header=0,low_memory=False)
        cellsIndex = df_X.index.values

        print('Input data dimensions is '+ str(df_X.shape))
        
        ########################################################
        # Prepare data for the KNN. Specify training and test sets
        # And standarized based on parameters from the training set ( is that contamination?)
        
        #train_set,test_set = train_test_split(df_X,test_size=0.8,random_state=42)
        full_set = df_X.values
        
        # The following standardizes but also converts to NUMPY
        # Not required as Xtilde was already scaled
        #scaler = StandardScaler()  
        #scaler.fit(full_set)
        #full_set = scaler.transform(full_set)
        
        #######################################################
        # Choice of algorithm: algorithm='brute' not good for large data sets
        # Returns distances,indices of rank nxk
        # k+1 because we want k neighbors besides the target point not including it
        # Takes about 5 hours using the whole set on a single core
        
        cluster_set=full_set
        #cluster_set = train_set

        print('Cluster: n_jobs is set to '+str(self.njobs))
        print('Dims of data set to be cluster fit '+str(cluster_set.shape))
        t0 = tm.time()
        #nbrs = NearestNeighbors(n_neighbors=k+1, algorithm='ball_tree',n_jobs=self.njobs).fit(cluster_set)
        #### Does contain the target index when set is FULL maybe not if it is not full
        print('Cluster using the DF data objects')

        nbrs = NearestNeighbors(n_neighbors=k+1, algorithm='ball_tree',n_jobs=self.njobs).fit(cluster_set)
        distances, indices = nbrs.kneighbors(full_set)
        # nbrs.kneighbors_graph(full_set).toarray()
        print('Time to process the clustering is '+str(tm.time()-t0))
        
        # NOTE: is you request the neighbors for a point in the training set that point will be returned even though it overlaps
        # print(neigh.kneighbors([[1., 1., 1.]])) 
        
        # NOTE I have demonstrated that sometimes the order is not what we want. Viz., the target is 
        # NOT always first. Here are a few lines from a indices file:
        #   these are snippets from a text file. NOTE the duplicated rows The first 2549 should be 2839
        # 2840 MB05_CTTCGCGCTAAA 2838 2824 3029 2867 2791 2722 2421 2716 2685 2671 2775
        # 2841 MB05_TGCTAACAGCGA 2549 2839 3035 2757 2872 2781 2898 2771 2525 2599 2728
        # 2550 MB05_GCCTAAATGCAC 2548 3049 2937 2599 2525 2771 2872 2898 2479 2757 3035
        # 2551 MB05_AGTGGTGGCGAA 2549 2839 3035 2757 2872 2781 2898 2771 2525 2599 2728
        # 2552 MB05_CCCAGGGGAACG 2550 2249 2379 2658 2285 2539 2400 2474 2373 2168 2357
        #
        # Construct header line
        headers = constructHeaderNames(k+1)
        df_indices = pd.DataFrame(indices)
        df_indices['CELL'] = cellsIndex[df_indices.index.values] # Assumes implied order is the cell order plus 
        df_indices.set_index('CELL',inplace=True)
        df_indices.columns=headers[1:]
        
        df_distances = pd.DataFrame(distances)
        df_distances['CELL'] = cellsIndex[df_distances.index.values]
        df_distances.set_index('CELL',inplace=True)
        df_distances.columns=headers[1:]

        self.outputIndices = df_indices
        self.outputDistances = df_distances

        ########################################################
        print('clustering procedure completed')



#
# Construct the data sets given the sizes n,p,k,and w
#
import pandas as pd
import time as tm
from broonie.generate_neighbor_matrix import generateNeighborMatrixClass
from broonie.construct_simulated_dataset import construct_simulated_dataset

metadata = 'SinglevectorsimulatedNew'
n=5000
p=200
k=10
w=2

simdat = construct_simulated_dataset(metadata,n,p,k,w)
simdat.constructDataObjects()
simdat.outputSimulationParams()
simdat.writeObjectsToDisk()




#################################################################
##
## Some elementary steps of the method by
## Aliverti, Lum, Johndrow, and Dunson.
## Attempt to build the simulation data set as instructed in 
## arXiv: 1810.08255v1 18,Oct,2018, pg 8
##
## READ preconstructed X and Z matrices
##
## NON iterative solution probably best for the p < n case. Solve
## The SVD equations directly

## Can we reintroduce the "linear" terms of group effects back into the Xtilde matrix?
## The nonlinear terms are lost to the residuals vector and so prob canot be recovered.

##################################################################

import pandas as pd
import numpy as np
from numpy import zeros
from numpy import diag
import math
import matplotlib.pyplot as plt
import sklearn.preprocessing
from scipy.sparse.linalg import svds, eigs
from scipy.linalg import svd
from scipy.optimize import line_search,brentq,brent
from scipy import stats
from sklearn import linear_model
from scipy.spatial.distance import cdist
from sklearn.model_selection import train_test_split
from sklearn.linear_model import Ridge
from sklearn.model_selection import KFold
from sklearn.model_selection import cross_val_score
from sklearn.linear_model import RidgeCV
from sklearn.datasets import load_boston
from sklearn.preprocessing import StandardScaler

import random
import sys

#############################################################################

#
# Also want to return VT,S for subset of betas set to ZERO 

class denseClassCovariateProjectionReductionReaddata(object):
    """Perform the projection of Z from X. Ultimately construct the final U and S
    matrices required to buld a new X-tilde matrix
    
    Attributes
    ----------
    self.metadata = 'base naming for output files U and S and Xtilde'
    self.X: input X array in numpy format
    self.Z: input Z array in nunpy format
    self.n: Number of obsewrvations
    self.p: Number of parameters
    self.k: number of components - an input parameter for variable reduction selection
    self.w: Number of Z covariates to remove (not tested sufficiently beyond 1)

    OUTPUTS VT, S (normalized)
    """

    def __init__(self, k, metadata='defaultName',initialGuess=False,inU=0,inS=0):
        self.metadata = metadata
        self.k = k
        self.betaIndicesToZero = list()
        self.prog_params = {
            'verbose':False,
            'debug':False }

    def set_prog_params(self, in_params):
        self.prog_params = in_params
        return

    def set_k(self, ink):
        self.k = ink
        return

    def get_X(self):
        return self.X

    def get_Z(self):
        return self.Z

    def specifyBetasToZero(self, inBetas):
        self.betaIndicesToZero = inBetas

    def readXandZdata(self, filenameX='XdataSimulatedAliverti.tsv', filenameZ='ZdataSimulatedAliverti.tsv'):
        dfX = pd.read_table(filenameX, delim_whitespace=True)
        self.X = dfX.values
        dfZ = pd.read_table(filenameZ, delim_whitespace=True)
        self.Z = dfZ.values
        self.n = self.X.shape[0]
        self.p = self.X.shape[1]
        self.w = self.Z.shape[1]
        self.data_read=True
        return

    def setXandYdata(self, inX,inZ):
        """X and Z are in numpy ndarray format
        """
        self.X = inX
        self.Z = inZ
        self.n = self.X.shape[0]
        self.p = self.X.shape[1]
        self.w = self.Z.shape[1]
        self.read_data = True
        return

    def outputMatrixRanks(self):
        print('n,p,k,w are')
        print(str(self.n)+' '+str(self.p)+' '+str(self.k)+' '+str(self.w))
        return

    def returnSVDdata(self):
        return self.U,self.d,self.VT,self.effectCorrection

###############################################################################
# Basic Solver equations: OK these have been tested
# Options for computing the betas required for Z projection

    def fetchBetaNaturalEquations(self,X,Z):
        """This can be used to solve for a number of Z terms >1
        On return is the new normalized s
        Actuallly, if al the effects are onehotencoded (0,1) and
        since every row as a catagory, then the inverse should always
        be singular becasue SUM(onehot)i = Meani
        So we should either do a pseudoinvsser with throw a message OR
        solve the OLS using function calls
        """
        print('Natural Equation solution using pseudoInverse')
        w = Z.shape[1]
        newz = np.concatenate((np.ones((Z.shape[0],1)),Z),axis=1)
        invZ = self.inverseZtZ(newz)
        betanew = np.matmul(invZ, np.matmul(np.transpose(newz),X))
        return (betanew.tolist())

    def fetchBetaLinearRegression(self,X,Z):
        """Solve using linregress. Can only have one Z component
        returns the new normalized s. Best numerical stability
        """
        print('Linear Regression solver')
        reg = linear_model.LinearRegression()
        reg.fit(Z,X)
        coefs = reg.coef_
        intercept = reg.intercept_.reshape(reg.intercept_.shape[0],1)
        data = np.append(intercept,coefs,axis=1).T
        return (data.tolist())

    def fetchBetaLinearRegressionOld(self,X,Z):
        """Solve using linregress. Can only have one Z component
        returns the new normalized s. Best numerical stability
        """
        slope, intercept, r_value, p_value, std_err = stats.linregress(Z[:,0],X[:,0])
        debug = self.prog_params['debug']
        if (debug):
            print('intercept '+str(intercept))
            print('slope '+str(slope))
        return ([[intercept],[slope]])

    def fetchBeta(self, X, Z, iflag=0):
        """interface to real solvers. Manually choose one for now
        iflag=0 linearregression solver
        iflag=1 natural equation solvers ( possible numerical instability)
        """
        if (iflag !=0 and iflag!=1):
            print('Wrong iflag choice: changing to iflag=0 linear regression solver')
            iflag=0
        if (iflag==0):
            return(self.fetchBetaLinearRegression(X,Z))
        else:
            return(self.fetchBetaNaturalEquations(X,Z))

###############################################################################
# Alternative Solver and related equations: 
# Options for computing the betas required for Z projection withj colinearity. Use ridge regularization
# Need a way to search out the best lambda terms
# Featuresa MUST be standardize here for proper training

    def fetchRidgeBeta(self, X, Z): 
        """Solve using ridge regression and simple CV determination of lambda
        Z here acts as the independant variable and X the dependent term
        NOTE inputs are PANDAS dataframes
        """
        alphaList =  [x/10.0 for x in range(1,100,5)]
        regr_cv = RidgeCV(alphas=alphaList)  # Provide a range of alphas to test
        model_cv = regr_cv.fit(Z, X)
        print ('Best model params is ='+str(model_cv.alpha_))
        #ridge = linear_model.Ridge(alpha=a, normalize=True,fit_intercept=False)
        #ridge.fit(X, y)
        #coefs.append(ridge.coef_)
        type(regr_cv.coef_)
        print(str(regr_cv.coef_.shape))
        betas = regr_cv.coef_[alphaList.index(model_cv.alpha_),:]
        intercept = regr_cv.intercept_[alphaList.index(model_cv.alpha_)]
        print('intercept '+str(intercept))
        print('slope '+str(betas))
        return ([[intercept],betas])
        print(results)

############################################################################
# MISC
    def checkUnormalization(self, U):
        """ Return the average error per diagonal of the input array
        """
        n = U.shape[0]
        s = np.sum(np.diag(np.matmul(np.transpose(U),U)))
        return ((n-s)/n)

    def inverseZtZ(self, Z):
        """return a 2D array pseudo Inverse 
        By construction this should always be singular
        """
        print('Using a Moore-Penrose pseudoinverse to solve the OLS')
        A = np.matmul(np.transpose(Z),Z)
        Ainv = np.linalg.pinv(A)
        return (Ainv)

    def centerAndScale(X):
        outX = sklearn.preprocessing.scale(Xprescaled,axis=0,with_mean=True,with_std=True, copy=True)
        return(outX)

    def normMatrix(self,M):
        """Simply return the normalized term for M
        IF the norm is ZERO then thorugh a warning and return M 
        unchanged
        """
        invM = np.linalg.norm(M)
        if (invM == 0.0):
            print('matrix norm is ZERO return M unchanged')
            return (M)
        return (M / invM )

###################################################################################
# Get the work done 
    def computeUandS(self):
        """Perform the full procedure here. We should refactor this but who has time now?
        """
        self.outputMatrixRanks() # This ensures handling the guess correctly by resetting class flags
        k = self.k
        n = self.n
        p = self.p
        w = self.w
        X = self.X
        Z = self.Z
        p = X.shape[1]

## Perform the SVD. 
## It is unclear if simply sorting in reverse is sufficient given the ridiculous 
## definition provided by scipy of sorting "not in increasing order"
## So we have to do this the hard way  - such madness
## Can choose to zero out betas (temp) here
        if (k==p):
            U, s, VT = svd(X.astype('d'))
        else:
            U, s, VT = svds(X.astype('d'),k=k,which='LM')
        high2low = np.argsort(s)[::-1][:k]
        s = s[high2low]
        U = U[:,high2low]
        VT = VT[high2low,:]
        d = np.identity(k)
        np.fill_diagonal(d, s)
        # Keep the U,S,VT for inverse OG transformations
        self.U = U
        self.VT = VT
        self.d = d
        lmX = np.matmul(U,d) 
        temp = self.fetchBeta(lmX,Z,iflag=0)
        self.beta = temp
        newz = np.concatenate((np.ones((Z.shape[0],1)),Z),axis=1)
        self.effectCorrection = np.matmul(newz , temp)
        S = lmX - self.effectCorrection
        return(VT,S)

    def computeUandS_ZscoredEigenvalues(self):
        """Select order of components based on the variance_explained Zscores rather than
        eigenvalue magnitude. It seems no difference in the selection
        """
        self.outputMatrixRanks() # This ensures handling the guess correctly by resetting class flags
        k = self.k
        n = self.n
        p = self.p
        w = self.w
        X = self.X
        Z = self.Z
        p = X.shape[1]
        if (k==p):
            U, s, VT = svd(X.astype('d'))
        else:
            U, s, VT = svds(X.astype('d'),k=k,which='LM')
        sdev = s / np.sqrt(np.maximum(1, X.shape[1] - 1))
        Proportion = np.square(sdev) / sum(np.square(sdev))
        m = Proportion.mean()
        sd = Proportion.std()
        zscore = (Proportion - m)/sd
        high2low = np.argsort(zscore)[::-1][:k]
        s = s[high2low]
        U = U[:,high2low]
        VT = VT[high2low,:]
        d = np.identity(k)
        np.fill_diagonal(d, s)
        # Keep the U,S,VT for inverse OG transformations
        self.U = U
        self.VT = VT
        self.d = d
        lmX = np.matmul(U,d)
        temp = self.fetchBeta(lmX,Z,iflag=0)
        self.beta = temp
        newz = np.concatenate((np.ones((Z.shape[0],1)),Z),axis=1)
        self.effectCorrection = np.matmul(newz , temp)
        S = lmX - self.effectCorrection
        return(VT,S)

    def computeUandS_ZscoredEigenvaluesReturnReconstructed(self,betaZeroList):
        """Select order of components based on the variance_explained Zscores rather than
        eigenvalue magnitude. It seems no difference in the selection
        Also return the Matricers that result from zeroing out the betas listed
        in betaZeroList
        """
        self.outputMatrixRanks() # This ensures handling the guess correctly by resetting class flags
        k = self.k
        n = self.n
        p = self.p
        w = self.w
        X = self.X
        Z = self.Z
        p = X.shape[1]
        if (k==p):
            U, s, VT = svd(X)
        else:
            U, s, VT = svds(X,k=k,which='LM')
        sdev = s / np.sqrt(np.maximum(1, X.shape[1] - 1))
        Proportion = np.square(sdev) / sum(np.square(sdev))
        m = Proportion.mean()
        sd = Proportion.std()
        zscore = (Proportion - m)/sd
        high2low = np.argsort(zscore)[::-1][:k]
        s = s[high2low]
        U = U[:,high2low]
        VT = VT[high2low,:]
        d = np.identity(k)
        np.fill_diagonal(d, s)
        lmX = np.matmul(U,d)
        temp = self.fetchBeta(lmX,Z,iflag=0) # Might want to regularized
        tempZeros = temp.copy()
        zerorow = [0.0]*k
        for i in self.betaIndicesToZero: # Set values to zero FOR EACH K component
            print(' Zero index includes intercept term '+str(i+1))
            tempZeros[i+1]=zerorow
        newz = np.concatenate((np.ones((Z.shape[0],1)),Z),axis=1)
        #print(str(temp))
        #print('_____________________________________________________________')
        #print('_____________________________________________________________')
        #print(str(tempZeros))
        S = lmX - np.matmul(newz,temp)
        self.effectCorrectionReconstruction = np.matmul(newz,tempZeros)
        SZeros = lmX - self.effectCorrectionReconstruction 
        # NOTE VT here corresponds to U in the other solvers
        return(VT,S,SZeros)

# Use the Ridge methods: THis might be required as beta values are rather big

    def computeUandS_ZscoredEigenvaluesRidge(self):
        """Select order of components based on the variance_explained Zscores rather than
        eigenvalue magnitude. It seems no difference in the selection
        """
        self.outputMatrixRanks() # This ensures handling the guess correctly by resetting class flags
        k = self.k
        n = self.n
        p = self.p
        w = self.w
        X = self.X
        Z = self.Z
        p = X.shape[1]
        if (k==p):
            U, s, VT = svd(X)
        else:
            U, s, VT = svds(X,k=k,which='LM')
        sdev = s / np.sqrt(np.maximum(1, X.shape[1] - 1))
        Proportion = np.square(sdev) / sum(np.square(sdev))
        m = Proportion.mean()
        sd = Proportion.std()
        zscore = (Proportion - m)/sd
        high2low = np.argsort(zscore)[::-1][:k]
        s = s[high2low]
        U = U[:,high2low]
        VT = VT[high2low,:]
        d = np.identity(k)
        np.fill_diagonal(d, s)
        lmX = np.matmul(U,d)
        temp = self.fetchRidgeBeta(lmX,Z)
        newz = np.concatenate((np.ones((Z.shape[0],1)),Z),axis=1)
        S = lmX - np.matmul(newz , temp)
        return(VT,S)




#################################################################
##
## Some elementary steps of the method by
## Aliverti, Lum, Johndrow, and Dunson.
## Attempt to build the simulation data set as instructed in 
## arXiv: 1810.08255v1 18,Oct,2018, pg 8
##
## READ preconstructed X and Z matrices
##
## Return the proper U and S matrices to construct Xtilde
## Xtilde is retunred as a NUMPY array
##
## The below code is equivelent to ALiverti et. al sSVD method
## sSVD is basically the same as sOG with the exclusion of the linear fit
##
##################################################################

import pandas as pd
import numpy as np
from numpy import zeros
from numpy import diag
import math
import matplotlib.pyplot as plt
import sklearn.preprocessing
from scipy.optimize import line_search,brentq,brent
from scipy import stats
import random

#####################################################################

class sSVDCovariateProjection(object):
    """Perform the projection of Z from X. Ultimately construct the final U and S
    matrices required ot buld a new X-tilde matrix
    
    Attributes
    ----------
    self.metadata = 'base naming for output files U and S and Xtilde'
    self.X: input X array in numpy format
    self.Z: input Z array in nunpy format
    self.n: Number of obsewrvations
    self.p: Number of parameters
    self.k: number of components - an input parameter for variable reduction selection
    self.w: Number of Z covariates to remove (not tested sufficiently beyond 1)
    self.conv_steps: Parameters driuving the convergence of Js
    self.conv_steps: Parameters driving the convergence of Theta (large u)
    self.prog_params: specify verbosity for now
    self.initialGuess declares intent to provuide a U and S starting matrix FROM which
        u and s initial vectors will be grabbed

    OUTPUTS U, S, (normalized),finalError list
    """

    def __init__(self, k, metadata='defaultName',initialGuess=False,inU=0,inS=0):
        self.metadata = metadata
        self.prog_params = {
            'verbose':False,
            'debug':False }
        self.conv_steps = {
            'lim_max':200,
            'lim_std':60,
            'lim_step':5 }
        self.conv_params = {
            'maxiter':500,
            'tolerance':1e-6,
            't_val':10,
            'diffTol':1e-7 }
        self.k = k
        self.data_read = False
        self.completeErrors = list()
        self.initialGuess=initialGuess
        self.guessU = inU
        self.guessS = inS

    def set_conv_params(self, in_params):
        self.conv_params = in_params
        return

    def set_conv_steps(self, in_steps):
        self.conv_steps = in_steps
        return

    def set_prog_params(self, in_params):
        self.prog_params = in_params
        return
 
    def set_k(self, ink):
        self.k = ink
        return

    def get_X(self):
        return self.X

    def get_Z(self):
        return self.Z

    def get_completeErrors():
        return (completeErrors)

    def readXandZdata(self, filenameX='XdataSimulatedAliverti.tsv', filenameZ='ZdataSimulatedAliverti.tsv'):
        dfX = pd.read_table(filenameX, delim_whitespace=True)
        self.X = dfX.values
        dfZ = pd.read_table(filenameZ, delim_whitespace=True)
        self.Z = dfZ.values
        self.n = self.X.shape[0]
        self.p = self.X.shape[1]
        self.w = self.Z.shape[1]
        self.data_read=True
        return

    def setXandYdata(self, inX,inZ):
        """X and Z are in numpy ndarray format
        """
        self.X = inX
        self.Z = inZ
        self.n = self.X.shape[0]
        self.p = self.X.shape[1]
        self.w = self.Z.shape[1]
        self.read_data = True
        return

    def outputMatrixRanks(self):
        print('n,p,k,w are')
        print(str(self.n)+' '+str(self.p)+' '+str(self.k)+' '+str(self.w))
        if (self.initialGuess):
            print('Initial guess for U and S is provided')
            inn=self.guessS.shape[0]
            inp=self.guessU.shape[1]
            ink=self.guessS.shape[1]
            print('Parameters found for initial Guesses: '+str(inn)+' '+str(inp)+' '+str(ink))
            if ( self.n != inn or self.p != inp or self.k != ink):
                print(self.guessU.shape)
                print(self.guessS.shape)
                print('Guess sizes mismatch n,p,or k: Reset guess status to RANDOM')
                self.initialGuess=False
        return

##################################################################
# Find Theta computations: These are usually inside of iterative search methods
# methods for finding the best THETA: Primarily when > t_val
# NOTE if X is not centered/scaled. this function gets heavily impacted versus t_val = 10 

    def Stheta(self, x, Xts, t_val=0):
        """Evaluate the function S_theta == S_theta(Xts) = =sign(Xts)(avbs(Xts)-theta)I(abs(Zts))
        """
        newX = self.SthetaMatrix(x, Xts)
        stheta = np.abs(newX).sum()-t_val
        if (self.prog_params['debug']): print('Stheta: input lambda is '+str(x)+' output Stheta estimate '+str(stheta))
        return(stheta)

# Apply the final Stheta to generate the U matrix
# In this situation "*" is a Hadamard product (element wise)
# Danger checking a float for zero

    def SthetaMatrix(self, theta, Xts):
        """Evaluate the function S_theta == S_theta(Xts) = =sign(Xts)(avbs(Xts)-theta)I(abs(Zts))
        If theta enters as None then no root NOR minimum was found. So simply return the original Xts vector
        as if the large phase iterations were never entered
        """
        if (theta is None):  # Should almost never get here
            print('Warning: Stheta was None should not commonly be here')
            #Xts / np.linalg.norm(Xts)
            return ( self.normMatrix(Xts))
        absXts = np.abs(Xts)
        signXts = np.sign(Xts)
        newX = signXts * (absXts-theta) * ((np.sign(absXts-theta)+1.0)/2.0)
        normNewX = np.linalg.norm(newX)
        if (normNewX==0.0):
            if (self.prog_params['debug']): print('Xnew norm is zero: probably root finder is at boundary')
            return (np.abs(newX)) # returns an empty vector of zeros
        else:
            return (newX / normNewX )

###############################################################################
# Solver equations
# Options for computing the betas required for Z projection

    def solveBetaNaturalEquations(self, s,Z):
        """This can be used to solve for a number of Z terms >1
        On return is the new normalized s
        """
        newz = np.concatenate((np.ones((Z.shape[0],1)),Z),axis=1) 
        invZ = self.inverseZtZ(newz)
        betanew = np.matmul(invZ, np.matmul(np.transpose(newz),s))
        s = s - np.matmul(newz, betanew)
        #s = s / np.linalg.norm(s)
        s = self.normMatrix(s)
        return (s)

    def solveBetaLinearRegression(self, s,Z):
        """Solve using linregress. Can only have one Z component
        returns the new normalized s. Best numerical stability
        """
        #slope, intercept, r_value, p_value, std_err = stats.linregress(Z[:,0],s[:,0])
        print('Linear Regression solver')
        reg = linear_model.LinearRegression()
        ref.fit(Z,s)
        coefs = reg.coef_
        intercept = reg.intercept_.reshape(reg.intercept_.shape[0],1)
        data = np.append(intercept,coefs,axis=1).T
        newz = np.concatenate((np.ones((Z.shape[0],1)),Z),axis=1)
        s = s - np.matmul(newz, data)
        return (s)

    def solveBeta(self, s, Z, iflag=0):
        """interface to real solvers. Manually choose one for now
        iflag=0 linearregression solver
        iflag=1 natural equation solvers ( possible numerical instability)
        """
        if (iflag !=0 and iflag!=1):
            print('Wrong iflag choice: changing to to iflag=1 natural eqn solvers')
            iflag=1
        if (iflag==0):
            return(self.solveBetaLinearRegression(s,Z))
        else:
            return(self.solveBetaNaturalEquations(s,Z))
        
#################################################################################
# iterative methods

    def runBrentQ(self, lim, Xts, t_val):
        """Run a single instance of BrentQ
        """
        try:
              thetaStar = brentq(self.Stheta, 0, lim, args=(Xts,t_val))
        except:
              thetaStar = None
        return (thetaStar)

    def iterativeRootFind(self, increased_lim, Xts, t_val):
        """Find ROOT of the Stheta function
        """
        debug = self.prog_params['debug']
        verbose = self.prog_params['verbose']
        if (debug): print('Enter iterativeRootFind')
        icount=0
        while (increased_lim < self.conv_steps['lim_max']):
              icount += 1
              thetaStar= self.runBrentQ(increased_lim, Xts, t_val)
              increased_lim += self.conv_steps['lim_step']
              if (verbose) :print('---- iterating over find root theta step lim is '+str(increased_lim)+' theta '+str(thetaStar))
        return(thetaStar,icount)

    def iterativeMinimumFind(self, increased_lim, Xts, t_val):
        """Find the MINIMUM of the Stheta function IFF no ROOT found
        """
        debug = self.prog_params['debug']
        verbose = self.prog_params['verbose']
        if (debug): print('Enter iterativeMinimizeFind')
        thetaStar = None
        try:
              thetaStar = brent(self.Stheta,args=(Xts,self.conv_params['t_val']))
        except:
              thetaStar = None
              if (verbose): print('---- Looking for minumum. lim is '+str(increased_lim)+' theta '+str(thetaStar))
        return(thetaStar)

################################################################################
# Misc methods

    def checkUnormalization(self, U):
        """ Return the average error per diagonal of the input array
        """
        n = U.shape[0]
        s = np.sum(np.diag(np.matmul(np.transpose(U),U)))
        return ((n-s)/n)
    
    def inverseZtZ(self, Z):
        """return a 2D array Inverse of the overlap matrix for the giuven input array
        """
        A = np.matmul(np.transpose(Z),Z)
        return (np.linalg.pinv(A))

    def centerAndScale(X):
        outX = sklearn.preprocessing.scale(Xprescaled,axis=0, with_mean=True, with_std=True, copy=True)
        return(outX)

    def initialGuessRandomU(self):
        """Abstract it for now since we may want to read in guesses from 
        other sources
        """
        u0 = np.random.normal(0,1,size = (self.p,1))
        #u0 = u0 / np.linalg.norm(u0)
        u0 = self.normMatrix(u0)
        return (u0)

    def initialGuessRandomS(self):
        """Abstract it for now since we may want to read in guesses from 
        other sources
        """
        s0 = np.random.normal(0,1,size = (self.n,1))  # Go ahead and keep s,u as this allocates memory for them 
        #s0 = s0 / np.linalg.norm(s0)
        s0 = self.normMatrix(s0)
        return (s0)

    def fetchGuesses(self,j):
        """Simply grab the jth entry from U and S and return as a vector u,s
        We don't need ot check errors too much as we've already confirmed
        U and S conform
        """
        print('get U and S guess for j = '+str(j))
        u = self.guessU[j,]
        s = self.guessS[:,j]
        #u.reshape(self.guessU.shape[1],1)
        #s.reshape(self.guessS.shape[0],1)
        return( u.reshape(self.guessU.shape[1],1),s.reshape(self.guessS.shape[0],1))
        return (u,s)

    def normMatrix(self,M):
        """Simply return the normalized term for M
        IF the norm is ZERO then thorugh a warning and return M 
        unchanged
        """
        invM = np.linalg.norm(M)
        if (invM == 0.0):
            print('matrix norm is ZERO return M unchanged')
            return (M)
        return (M / invM )

################################################################################
# Perform the full solution set here
# Begin algorithym: Specify K the "cluster size" from which projections will be performed
# On input we need X(nxp) and Z(nx1). Information regarding the error convergence rates
# are collected for future analyses

    def computeUandS(self):
        """Perform the full procedure here. We should refactor this but who has time now?
        """
        self.outputMatrixRanks() # This guarentees handling the guess correctly by resetting class flags
        k = self.k
        n = self.n
        p = self.p
        w = self.w
        prog_params = self.prog_params
        conv_params = self.conv_params
        conv_steps = self.conv_steps
        X = self.X
        Z = self.Z
# Specify Initial guesses: is reading in a guess potentially useful?
        U = np.zeros((k,p)) 
        S = np.zeros((n,k))
        P = np.zeros((n,n))
        u0 = self.initialGuessRandomU() # May be used even if you specified a guess U
        s0 = self.initialGuessRandomS()
        convergedJs = ['Notconverged']*k
        finalError = [-99]*k
        #completeErrors = list() 
        verbose = prog_params['verbose']
        for j in range(0,k):
            if (verbose): print('Perform outer sweep for j = '+str(j))
            nit = 0
            err = 10
            #
            if (self.initialGuess):
                u,s = self.fetchGuesses(j)  # NOTE we do not update the guess data Just use as a starting point
            else:
                u = u0
                s = s0
            px = np.matmul(np.identity(n)-P,X) if j != 0 else X
            error = list()
            deltaErr=err
            while (nit < conv_params['maxiter'] and err > conv_params['tolerance'] and deltaErr > conv_params['diffTol']):
                if (verbose): print('While iteration '+str(nit)+' for J = '+str(j))
                errold = err
                sold = s
                uold = u
                s = np.matmul(px,u)  
                s = self.normMatrix(s)
                s = s / np.linalg.norm(s)
                ##Not used s = self.solveBeta(s,Z,iflag=0) #iflag 0 is linregress 
                Xts = np.matmul(np.transpose(X),s)
                if (np.abs(Xts).sum() < conv_params['t_val']):
                    print('small u < t_val')
                    #u = Xts / np.linalg.norm(Xts)
                    u = self.normMatrix(Xts)
                else:
                    if (verbose): print('large u: >= t_val'+str(conv_steps['lim_std'])+' '+str(conv_params['t_val']))
                    thetaStar = self.runBrentQ(conv_steps['lim_std'], Xts, conv_params['t_val'])
                    increased_lim = conv_steps['lim_std'] + conv_steps['lim_step']
                    if (thetaStar is None):
                        thetaStar,icount = self.iterativeRootFind(increased_lim, Xts, conv_params['t_val'])
                    if (thetaStar is None): # try another method to estimate thetaStar
                        print('Find theta using minimizer')
                        thetaStar = self.iterativeMinimumFind(increased_lim, Xts, conv_params['t_val'])
                        print('minimum stheta is  '+str(thetaStar))
                    u = self.SthetaMatrix(thetaStar, Xts)
                nit += 1
                err = (sum(np.abs(s-sold)) + sum(np.abs(u-uold)))[0]
                if (err <= conv_params['tolerance']):
                    print('U: J term '+str(j)+' converged to '+str(err)+' in numiter = '+str(nit))
                    convergedJs[j]='Converged'

                error.append(err)    # Grab these to study convergence behavior for a given j
                finalError[j] = err
                deltaErr = np.abs(errold-err) # alternative criteria for completing J
                if (deltaErr <= conv_params['diffTol'] and convergedJs[j] != 'Converged'):
                    print('U: J term '+str(j)+' stuck at '+str(err)+' in numiter = '+str(nit))
                    if (verbose): print('Warning (u/s) is stuck')
                    convergedJs[j]='Stuck'
            P = P + np.matmul(s,np.transpose(s))
            U[j,] = u.flatten()
            S[:,j] = s.flatten()
            self.completeErrors.append(error)
            if (verbose):
                print('--------- Status of converged for each J for K = '+str(k))
                print(convergedJs)
                print(finalError)
        d = np.identity(k)
        np.fill_diagonal(d, diag(np.matmul(np.transpose(S),np.matmul(X,np.transpose(U)))))
        print(diag(d))
        S = np.matmul(S,d)
        print('Finished projecting Z out of Xk')
        #Xtilderaw = np.matmul(S,U)
        #Xtilde =  sklearn.preprocessing.scale(Xtilderaw,axis=0, with_mean=True, with_std=True, copy=True)
        print('Final convergence status')
        print(convergedJs)
        print(finalError)
        return(U,S,finalError)



#################################################################
##
## These are the construction of the data from Ben using Zscores.
## SO we do not filter by zeros. but we continue to use MAD() thresholding

## BUILD a new datafrae as follows. Columns CELLS MOUSE SEX DATE VEHICLE gene1 gene2 ....genep
## The gene values should be PRESCALED but not whitened. We can do thatlater
## The MOUSE value can be constructed form the CELL name

# MetaDataWithDate
##"nGene" "nUMI"  "Replicate"     "Treatment"     "Sex"   "Date"  "ID"

## So we have initially 16,600 Genes and 29,538 SingleCells
##
##################################################################

import time as tm
import pandas as pd
import numpy as np
import collections
import matplotlib.pyplot as plt
import math 
import scipy 
from scipy import stats
import sys
import sklearn.preprocessing

import sys
from numpy.random import seed
from numpy.random import randn
from statsmodels.graphics.gofplots import qqplot
import pylab
import scipy.stats as stats
from scipy import linalg

#from __future__ import print_function
#########################################################################

# Change these to suit
# Not all these files are always used

# Input range of cell types classified as TUMORS to INCLUDE
#if len(sys.argv) >2:
#    lowValue = int(sys.argv[1])
#    highValue = int(sys.argv[2])
#elif (len(sys.argv) == 2):
#    lowValue = int(sys.argv[1])
#    highValue = int(sys.argv[1])
#else:
#    lowValue = 1
#    highValue = 10
#print("Range of cells to keep is "+str(lowValue)+' to '+str(highValue))
#if (highValue > 10):
#    print("High Value may be too high should be 10 or less "+str(highValue))
#if (lowValue < 1):
#    print("Low Value may be too low should be 1 or greater "+str(lowValue))

rootdir = '/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/'
#regressedfilename = rootdir+'regressedResults/GT.scale.data.tsv'
regressedfilename = rootdir+'regressedResults/ScaleDataMatrix.tsv'
#metafilename = rootdir+'rawResults/MetaDataMatrix.txt'
metafilenameDate = rootdir+'regressedResults/MetaDataMatrixWithDate.txt'

t0=tm.time()
data = pd.read_csv(regressedfilename,delim_whitespace=True)
print(tm.time()-t0)

##
## Some basic plots of the regressed Zscore imputed data
##

datameans = pd.DataFrame(data.mean(axis=1))
datameans.reset_index(inplace=True)
datameans.columns=['GENE','MEANS']
#datameans.plot(kind='line',x='GENE',y='MEANS',label='Zscore imputed data: Column MEANS of regressed scRNAseq data')
#datameans.xlabel=('Gene')
#plt.ylabel=('Mean')
#plt.savefig('imputedData_GeneMeanCounts.pdf')
#plt.show()

##
## How about the cell mean reads?
##
datacellmeans = pd.DataFrame(data.mean())
datacellmeans.reset_index(inplace=True)
datacellmeans.columns=['CELL','CELLMEANS']
datacellmeans.sort_values(by=['CELL'],inplace=True)
#datacellmeans.plot(kind='line',x='CELL',y='CELLMEANS',label='Column MEANS of Zscore imputed scRNAseq data')
#datacellmeans.xlabel=('Cell')
#datacellmeans.ylabel=('CellMean')
#plt.savefig('imputedData_CellMeanCounts.pdf')
#plt.show()

##
## How about the cell mean reads?
##

datacellmeans = pd.DataFrame(data.mean())
datacellmeans.reset_index(inplace=True)
datacellmeans.columns=['CELL','CELLMEANS']
datacellmeans.sort_values(by=['CELL'],inplace=True)
#datacellmeans.plot(kind='line',x='CELL',y='CELLMEANS',label='Column MEANS of imputed scRNAseq data')
#datacellmeans.xlabel=('Cell')
#datacellmeans.ylabel=('CellMean')
#plt.savefig('rawData_CellMeanCounts.pdf')
#plt.show()

######################################################################
## Compute gene MAD() values then keep 5000 top
## CHeck the profile to see how many to keep

datamad = pd.DataFrame(data.mad(axis=1))
datamad.reset_index(inplace=True)
datamad.columns=['GENE','MAD']
datamad.sort_values(by='MAD',ascending=False,axis=0,inplace=True)
#datamad.plot(kind='line',x='GENE',y='MAD',label='Kirk: Imputed data: MAD() of imputed scRNAseq data')
#plt.ylabel=('MAD')
#plt.savefig('imputedData_GeneMADCounts.pdf')
#plt.show()

## Sort data set  and include only the4 Top 5000 MAD values
topGenes = datamad.head(5000)
topGenesList = topGenes['GENE'].tolist()
pd.DataFrame(topGenesList).to_csv('topGenes_MAD.tsv')

# Kirk wants to keep all genes for now

######################################################################
## Get list of TUMOR names for future use and proportions
## Keep all cells

def tumor_within_cellName(tumorname,cellname):
    return tumorname in cellname

def computeSizeFactor_cellname(cellname, counterObject):
    name = cellname.split('_')[0]
    numbcells = float(counterObject[name])
    total = float(sum(counterObject.values()))
    return numbcells/total

#fulllistNames = list(data.loc[topGenesList].columns.values)
fulllistNames = list(data.columns.values)
unique_tumors = set([ x.split('_')[0] for x in fulllistNames ])
#{'MB10', 'MB13', 'MB05', 'MB14', 'MB06', 'MB04', 'MB15', 'MB07', 'MB12', 'MB09'}
allentries_tumors_temp = list([ x.split('_')[0] for x in fulllistNames ])
from collections import Counter
counter = collections.Counter(allentries_tumors_temp)

## 1) Get some base data: Compute total ave read count from the Raw data
## just for comparisons
## We assume missingness is not to0 bad. But it seems we should QC the raw data 
## to get the following values  (?)

## skip the scaling of the reads

#data_new = data.loc[topGenesList]
data_new = data

# Confirm total reads per cell are the same
cellsums = data_new.sum(axis=0) # Values of 21019

# What about total reads per gene
genesums = data_new.sum(axis=1)
#genesums.hist(bins=20,range=[0,100000])
#plt.savefig('imputedData_GeneSums_postScale.pdf')
#plt.show()

###############################################################################3
# No exclusons based on zeros

## Post QC Plot information
# Per gene
datameans = pd.DataFrame(data_new.mean(axis=1))
datameans.reset_index(inplace=True)
datameans.columns=['GENE','VARIANCE']
#datameans.plot(kind='line',x='GENE',y='VARIANCE',label='Imputed data: Column Scaled-MEANS of imputed scRNAseq data')
#datameans.xlabel=('Gene')
#datameans.ylabel=('Scaled-MEANS')
#plt.savefig('imputedData_GeneMeanCounts_removeMissingness.pdf')
#plt.show()

# Per Cell
datacellmeans = pd.DataFrame(data_new.mean())
datacellmeans.reset_index(inplace=True)
datacellmeans.columns=['CELL','CELLMEANS']
datacellmeans.sort_values(by=['CELL'],inplace=True)
#datacellmeans.plot(kind='line',x='CELL',y='CELLMEANS',label='Column MEANS of imputed scRNAseq data')
#datacellmeans.xlabel=('Cell')
#datacellmeans.ylabel=('CellMean')
#plt.savefig('rawData_CellMeanCounts_removeMissingness.pdf')
#plt.show()

############################################################################################
# cell reads for al lmice have now been QC'd and scaled. We start filtering away non-vehicle, non-tumor data
# Then build data into the dataframe to apture some batch effects such as Date and Sex

## First Remove all CELLs that are not actually part of a Tumor using the MAP from Kirk
## /proj/kirklab/users/KirkFiles/
## We also only want VEHICLE cells

### This whoe procedure has changed AGAIN

# NODENAMES THAT ARE NOT TUMORS: Remaining terms have the nomenclaturee NODE A, NODE B, etc
# Dec 12. Per kirk. Keep al cell types but exclude the vismodigib mice: onlty keep vehicle
# PLACE A DUMMY TERM INTO NOTTUMORS AND RERUN

#notTumors =['Endothelial Cells','Vascular Fibroblasts','Astrocytes','Endothelial','Microglia','Neurons','Oligodendrocytes','Vascular']
notTumors =['TEST']

metadata = pd.read_csv(metafilenameDate,delim_whitespace=True)

# Get list of cells that are NOT tumors
listNotTumorCells_df = metadata.ix[((metadata['ID'].isin(notTumors)))]
listNotTumorCells = list(listNotTumorCells_df.index.values)

## Now get list of cells from metadatawithdate that are NOT Vehicle for exclusion

## Speciy DUMMY to fetch all cells

#listNotVehicleCells_df = metadata.ix[ metadata['Treatment'] != 'Vehicle']
listNotVehicleCells_df = metadata.ix[ metadata['Treatment'] == 'Dummy']
listNotVehicleCdells = list(listNotVehicleCells_df.index.values)

# Union unique of exclude lists
excludeList = set(listNotTumorCells+listNotVehicleCdells)
data_new_tumor = data_new.copy()
data_new_tumor.drop(excludeList,axis=1,inplace=True)

newY = metadata['Treatment']
newY.drop(listNotVehicleCdells,inplace=True)
newY.to_csv('mouse_Ydata_allPhenotypes_regressed_data_vehicle_tumor_scaled.tsv',sep=" ")

cellNames = data_new_tumor.iloc[1].to_frame()
cellNamesList = cellNames.index.tolist()
geneNames = data_new_tumor.iloc[:,1].to_frame()
geneNamesList = geneNames.index.tolist()
geneNamesList_cell = geneNamesList

# get data from numpy and convert to df
# Keep only the top5000

# temp = data_new_tumor.T[topGenesList].copy()
# data_new_tumor = temp.T

#data_new_tumor_scaled_T = pd.DataFrame(sklearn.preprocessing.scale(data_new_tumor.T, with_mean=True, with_std=True, copy=True).tolist())

data_new_tumor_scaled_T = data_new_tumor.T
data_new_tumor_scaled_T['CELL'] = cellNamesList
data_new_tumor_scaled_T.set_index('CELL',inplace=True)
data_new_tumor_scaled_T.columns = geneNamesList_cell

# Add metadata to the scaled data object (SEX,DATE,MOUSE)

scaledData = data_new_tumor_scaled_T.copy()
newcolumn = metadata.loc[scaledData.index.values][['Treatment','Sex','Date','ID']]

##
## Splitting on spaces wil cause problems. Simply remove spaces and let downstream live with it
##newcolumn.ID = newcolumn.ID.str.split(' ').str.get(1)
newcolumn.ID = [x.replace(' ','') for x in list(newcolumn.ID)]

# Now let's build a column that only has a mouse ID. We can convert to multiple rows of 
# indicator variables at some later time

mouseIndex = [ int(x.split('_')[0][2:]) for x in newcolumn.index ]
newcolumn['MOUSE']=mouseIndex

# Build final dataframe for disk storage
combinedScaledData = pd.merge(newcolumn,scaledData,left_index=True,right_index=True)
combinedScaledData.rename(columns={'Treatment':'TREATMENT', 'Sex':'SEX','Date':'DATE','ID':'NODEID'},inplace=True)

outfilename = 'mouse_allPhenotypes_regressed_data_allCells_Notscaled.tsv'
combinedScaledData.to_csv(outfilename,header=True,sep=" ",index=True)
################################################################
# might as well build the whitened data set too

# NOTE here we skip adding the TREATMENT terms
# Fix below here
X = data_new_tumor_scaled_T.values
Carray = data_new_tumor_scaled_T.cov().values
D,V = linalg.eigh(Carray)
indices = D.argsort()[::-1]
D,V = D[indices], V[:,indices]
Dhalf = np.sqrt(D)
DhalfInv = 1.0/Dhalf
DmatrixHalfInv = np.diag(DhalfInv)
CarrayInvHalf = V @ DmatrixHalfInv @ V.T
whiteX3 =  X @ CarrayInvHalf
newCovar = np.cov(whiteX3,rowvar=False)

##
## Must create a new DataFrame by hand
##
newDF = pd.DataFrame(whiteX3)
newDF.columns = geneNamesList
newDF['CELL']=data_new_tumor_scaled_T.index.values
newDF.set_index('CELL',inplace=True)

##
## Now add the metadata again
##
# Add metadata to the scaled data object (SEX,DATE,MOUSE)

newcolumn = metadata.loc[newDF.index.values][['Sex','Date','ID']]
##newcolumn.ID = newcolumn.ID.str.split(' ').str.get(1)
newcolumn.ID = [x.replace(' ','') for x in list(newcolumn.ID)]

# Now let's build a column that only has a mouse ID. We can convert to multiple rows of 
# indicator variables at some later time

mouseIndex = [ int(x.split('_')[0][2:]) for x in newcolumn.index ]
newcolumn['MOUSE']=mouseIndex

# Build final dataframe for disk storage
combinedWhiteData = pd.merge(newcolumn,newDF,left_index=True,right_index=True)
combinedWhiteData.rename(columns={'Sex':'SEX','Date':'DATE','ID':'NODEID'},inplace=True)

outfilename = 'mouse_allPhenotypes_regressed_data_allCells_Notscaled_whitened.tsv'
combinedWhiteData.to_csv(outfilename,header=True,sep=" ",index=True)




#
# Construct the data sets given the sizes n,p,k,and w
# convert the covariate data to factors and/or indicator columns
# Jan 11:
#    Input argument for input file.
#    Modify list columnEffects to simply grab all THEN check for validity of input effect 
#
# New approach. Here we build a pipeline for imputing the RAW input data.
# We begin initially with the regressed data from BEN and so that data are also required.
# Add a method to return TWO matrices. One with all effects removesd and One with some effects 
# retained by zeroin out betas

# minMAx scaling as a possible alternatrive and allowing us to use Xtilde in the regression

# NOTE: Features are scaled (mean-centered or MinMax) but cells are not

###########################################################################

## Basic inpout file format
## CELL SEX DATE NODEID MOUSE 0610007N19Rik ........

import pandas as pd
import numpy as np
import time as tm
#import math
#import sys,ast
#import matplotlib.pyplot as plt
import collections
import sklearn.preprocessing
from sklearn import preprocessing
from sklearn.preprocessing import MinMaxScaler
from denseClassCovariateProjectionReductionReaddata import denseClassCovariateProjectionReductionReaddata

##########################################################################
# Basic functions

def constructMetaDataReadMatrix(indata,inColumnsForZ,batchfilename):
    """Determine what the possible set if EFFECTs are in the data set. Also updates
    a list of effects to treat as onehot encoded 
    """
    columnEffects,oneHotList = getKnownGroupEffects(indata,batchfilename)  # What effecs were found in fulldata
    # Which input exclusion effects remain valid  and were any left over ?
    columnsForZ = checkGroupEffectsValidity(inColumnsForZ, columnEffects)
    print('Input: Valid group covs are')
    print(columnsForZ)
    # First just get X the design data with no covariates information
    df_X = indata.drop(columnEffects,axis=1,inplace=False)
    # Need metadata for the final data object 
    indexing = df_X.index.values
    headers = df_X.columns.values
    #Now build the Z matrix  based on the input list.
    df_Z = indata[columnEffects]
    #Encode the Zs in preparation for x projection analysis
    df_Zall = df_Z 
    #print('oneHotList '+str(oneHotList))
    for effect in columnsForZ:
        #print('effect '+effect)
        if(effect in oneHotList):
            df_Zall = pd.merge(df_Zall, convertCovariateStringToOneHotEncoder(indata,effect),left_index=True,right_index=True)
        else:
            df_Zall = pd.merge(df_Zall, convertCovariateStringToFactors(indata,effect),left_index=True,right_index=True)
    df_Z = df_Zall.drop(columnEffects,axis=1,inplace=False)
    return df_X, df_Zall, df_Z, oneHotList
    
    # Special function when choosing MOUSE as a group variable
    # Convert a vector of 1s,2s,3s, etc. so several columns with 0s,1s exclusive
    # basically a set of onehot encoders
    # Just build and name a proper column of data then use sklearn oneHot encoding
    # Aliverti reported using this method doesn't have significant difference

    # TODO allow the USER to choice which effects should be onehot encoded
    
def convertCovariateStringToOneHotEncoder(df_in,effect):
    """We expect Z on entry to be an nx1 dataframe with the values
    In this approach we take the indicated colname and create new indicator 
    """
    # First encode the data
    indexing = df_in.index.values
    name = effect+'-encoded'
    df_Z = pd.DataFrame()
    df_Z = convertCovariateStringToFactors(df_in,effect)
    # Now expand and onehot
    enc = sklearn.preprocessing.OneHotEncoder(categories='auto')
    enc.fit(df_Z)
    onehotlabels = enc.transform(df_Z).toarray()
    data = onehotlabels
    numLabels = data.shape[1]
    newNameMeta = effect+'_ENC'
    newdf = list()
    for i in range(0,numLabels):
        namei = newNameMeta+str(i)
        newdf.append(pd.DataFrame({namei:data[:,i]}))
    df_onehot = pd.concat(newdf,axis=1)
    #df_onehot=pd.DataFrame({'MouseEnc0':data[:,0],'MouseEnc1':data[:,1],'MouseEnc2':data[:,2],'MouseEnc3':data[:,3],'MouseEnc4':data[:,4]})
    df_onehot['CELL']=indexing
    return (df_onehot.set_index('CELL'))

def convertCovariateStringToFactors(df,colname):
    """Convert Covariate factors to encoded integers
    Returns a new dataframe of one column
    """
    indexing = df.index.values
    newcolumnName = str(colname)+str('-encoded')
    le = sklearn.preprocessing.LabelEncoder()
    le.fit(df[colname].tolist())
    newCovariate = le.transform(df[colname].tolist())
    df_new = pd.DataFrame()
    df_new[newcolumnName] = newCovariate.tolist()
    df_new['CELL']=indexing 
    return (df_new.set_index('CELL'))

def getKnownGroupEffects(fulldata,batchfilename):
    """ Check the input file for KNOWN possible group effects
    Exclude those that may not exist such as TREATMENT. Make no
    assumptions about order or sorting. Return list in arbitrary 
    order. Independent of list provided by the user
    """
    foundColumns = list()
    listOneHotEncoded = list()
    if (batchfilename == False):
        print('Using prechosen batch effect names')
        currentPossibleColumnEffects = ['TREATMENT','MOUSE','SEX','DATE','NODEID']
        allColumns = fulldata.columns.values
        listOneHotEncoded = ['MOUSE'] # By default for use with our data testing
        # Grab only those fulldata columns that exist inb currentPossible
        foundColumns = list(set(allColumns) & set(currentPossibleColumnEffects))
    else: 
        foundColumnsCombined = list()
        with open(batchfilename) as f:
            tempColumns = f.read().splitlines()
            for effect in tempColumns:
                #foundColumnsCombined.append(effect.upper())
                foundColumnsCombined.append(effect)
            for ipair in foundColumnsCombined:
                #print('effect is '+ipair)
                name,typeEnc = ipair.strip().split(' ')
                foundColumns.append(name)
                if (typeEnc=='ONEHOT'):
                    listOneHotEncoded.append(name)
    print('Number of Possible batch effects found/specified is '+str(len(foundColumns)))
    print('Number of groupeffect columns in data is '+str(foundColumns))
    #print('They have been converted to UPPER CASE: '+str(foundColumns))
    return (foundColumns,listOneHotEncoded)

def checkGroupEffectsValidity(columnsForZ, columnEffects):
    """On input user request removal of columnsForZ
    Here we confimr if they all exist in the current data set. If not
    reduce columnsForZ to include only valid effects. Print out a message
    NOTE: always convert to and return uppercase for comparisons and subsequent analysis
    """
    upperColumnsForZ =list()
    for effect in columnsForZ:
        #upperColumnsForZ.append(effect.upper())
        upperColumnsForZ.append(effect)
    newList = list(set(upperColumnsForZ) & set(columnEffects))
    if ( newList.sort() != upperColumnsForZ.sort()):
        print('Input group list has been reduced based on preestablished possible effects')
        print('Size on data effects is '+str(len(columnsForZ))+' new removal list is of size '+str(len(newList)))
        print('New removal list is '+str(newList))
    return (newList)

def flatten(x):
    if isinstance(x, collections.Iterable):
        return [a for i in x for a in flatten(i)]
    else:
        return [x]

###############################################################################
# Start the class

class projectGroupDataClass(object):
    """ From the single-cell data reads. Perform a group-efforts projection using the OG method.
    Construct Xtilde and Z (covariates) Xtilde is standardized by default. Xtilde and Z are stored to disk and also 
    maybe returned using a class getter.

    Perform a group effects reduction of the sc data: 
    Supply a file of effects contained in your data set OR 
    by default anticipates covariates with the following names:
    ['TREATMENT','MOUSE','SEX','DATE','NODEID']
    On completion an Xtilde and associated Z matrix are written out
    On input:
    K = size of subspace for reduction 
    inputList = _ delimited list of covariates to EXCLUDE e.g., MOUSE_SEX_DATE
    infilename =  unscaled READs matrix including covariates
    outXtildemeta = metadata for building newfiles names for :
    outfilebase = outXtildemeta+str(K)
    Xtilde: outfilebase_Xtilde_Order.tsv'
    Z     : outfilebase_Z_Order.tsv'
    batchfilename point to a file that contains n rows and 1 column with relevant batch effect names:
    The list doesn't need to be exclusive. You do not need to remove all effects. Moreover, 
    For future version one can recover effect contributions as well

    Attributes
    ----------
    self.k = subspace from which projection will occur 
    self.inputList: UNDERSCORE delimited list of effects to project away. E.g., MOUSE_SEX_DATE
    self.batchfilename (opt) file that contains n rows and 2 column with relevant batch effect names
            col 1 effect name: col two either ONEHOT or CAT for type of inclusion
    self.scale = True/False. Standardize the Data prior to projection (default is True)
    self.outXtildemeta = String descriptor. Used to construct output filenames:
            outfilereconstructdata = outXtildemeta+'_reconstruct'+self.inReconstructList+'_'+str(K)
            self.outputZReconstructfile = outfilereconstructdata+'_Z_ZscoreOrder.tsv'
            self.outputXReconstructfile = outfilereconstructdata+'_Xtilde_ZscoreOrder.tsv'
    self.reconstructData (True/False). Declares building an additional Xtilde matrix keeping select effects
            in the matrix by zeroing out their betas
    self.inReconstructList: UNDERSCORE delimited list of effect to reintroduce if self.reconstruct Data=True
    self.origScaling True is mean-centered,std=1 scaling. False is MinMax scaling
    """
    def reportParameters(self):
        print('Infilename is '+self.infilename)
        print('Output metadata '+self.outXtildemeta)
        print('K subspace is '+str(self.k))
        print('exclude covariates list is '+self.inputList)
        print('Scale features is '+str(self.scale))
        if (self.reconstructData):
            print('A second Xtilde matrix will be created retaining the effects of:')
            print(self.inReconstructList)

    def __init__(self, k, infilename, inputList, outXtildemeta, batchfilename=False, scale=True, origScaling=True):
        self.k = k
        self.inputList = inputList
        self.infilename = infilename
        self.outXtildemeta = outXtildemeta
        self.batchfilename = batchfilename
        self.fulldata = pd.DataFrame()
        self.df_Xtilde = pd.DataFrame()
        self.df_XtildeZero = pd.DataFrame()
        self.df_Zall = pd.DataFrame()
        self.outputXfile = 'notSet'
        self.outputZfile= 'notSet'
        self.scale = scale if scale==False else True 
        self.reconstructData = False
        self.origScaling = origScaling if origScaling==False else True
        self.listOneHot = ['MOUSE'] # A default
        self.outputXReconstructfile = 'notset'
        self.outputZReconstructfile = 'notset'

    def setOrigScaling(self,inOrigScale):
        self.origScaling = inOrigScale if inOrigScale==False else True

    def reportScaling(self):
        print('Scaling Original type is '+str(self.origScaling))
   
    def fetchK(self):
        return(self.k)

    def returnSVDdata(self):
        return self.U,self.d,self.VT,self.effectCorrection

    def mapZtoBetas(self, inReconstructList,df_Z):
        """Given the input compound tab_delimited string, determine the beta indexing for removal
        This is tricky since some Zs are oneHot encoded giving rise to 
        multiple Betas. df_Zall already has all Z names in the headings
        and in order so use that. If the reconstructed inList doesn't exist
        set betas to -1 and reset self.reconstructData = False
        """
        #print('in MAP '+str(inReconstructList))
        covlist = inReconstructList.split('_')
        inColumnsForZ=list()
        for word in covlist:
            print('word '+word)
            inColumnsForZ.append(word)
        headers = df_Z.columns.values
        print(str(headers))
        print(str(inColumnsForZ))
        allindices = [i for i, s in enumerate(headers) if 'encoded' or 'ENC' in s]
        rawindices = [i for i, s in enumerate(headers) if ('encoded' in s) and (s.split('-')[0] in inColumnsForZ)]
        rawindices.append([i for i, s in enumerate(headers) if ('ENC' in s) and (s.split('_')[0] in inColumnsForZ)])
        #print('before reduction '+str(rawindices))
        minIndex = np.min(allindices)
        indices = [i-minIndex for i in flatten(rawindices)]
        indices.sort()
        print('sorted indices not including intercept term')
        return (indices)

    def reconstructList(self, inReconstructList):
        self.inReconstructList = inReconstructList
        self.reconstructData = True

    def fetchXtilde(self):
        return self.df_Xtilde
        
    def fetchZ(self):
        return self.df_Zall

    def fetchReconstructedXtilde(self):
        return self.df_XtildeZero

    def getXtildeFilename(self):
        return self.outputXfile
    
    def getZFilename(self):
        return self.outputZfile

    def getReconstructedXtilde(self):
        return self.outputXReconstructfile 

        ##########################################################
        # Start the work
        #
        # Specify as (for example) python denseProjectGroupData_MouseTumor.py 12 "DATE_MOUSE"
        # columnsForZ = ['MOUSE','SEX','DATE','NODEID']
        # Using a loop strcture to better control for column metadata
        # Example python projectGroupDataProcedure_MouseTumor.py 1000 'MOUSE_SEX_DATE'
        # inputList='SEX_DATE_MOUSE'
    
    def runProjection(self):
        K = self.k
        inputList = self.inputList
        infilename = self.infilename
        outXtildemeta = self.outXtildemeta
        covlist = inputList.split('_')
        inColumnsForZ=list()
        for word in covlist:
            inColumnsForZ.append(word)
        print('in K is '+str(K))
        print('Input: Unchecked group covs are')
        print(inColumnsForZ)
        print('Input: filename is '+infilename)
        print('Input: output Xtilde filename is '+outXtildemeta)
        if (self.reconstructData):
            print('Will also construct secondary Xtilde matrix zeroing out betas for')
            print(self.inReconstructList)
        # Metadata for the projection methods
        #metadata='mouseTumor_regressed'
        #runtype='EIGEN'
        metadata='mouseTumor_Zscoreregressed'
        runtype='ZSCORE' # Perform projectionn using Zscore ranked eigenvalues else uses eigenvalue magnetudes 
        # build outputfilename metadata
        outfilebase = outXtildemeta+str(K)
        outfilebase_meta=''
        outfilemetadata = outfilebase
        self.outputXfile = outfilemetadata+'_Xtilde_ZscoreOrder.tsv'
        self.outputZfile = outfilemetadata+'_Z_ZscoreOrder.tsv'
        if (self.reconstructData):
            outfilereconstructdata = outXtildemeta+'reconstruct'+self.inReconstructList+'_'+str(K)
            self.outputZReconstructfile = outfilereconstructdata+'_Z_ZscoreOrder.tsv'
            self.outputXReconstructfile = outfilereconstructdata+'_Xtilde_ZscoreOrder.tsv'
            print('Reconstructed X is '+self.outputXReconstructfile)
        self.fulldata = pd.read_table(infilename,delim_whitespace=True, index_col=0,header=0,low_memory=False)
        #########################################################
        # Build proper formatted data objects. Remove all Zs from fulldata and add to df_Z
        df_X, df_Zall, df_Z, self.listOneHot = constructMetaDataReadMatrix(self.fulldata,inColumnsForZ,self.batchfilename)
        self.df_Zall = df_Zall
        df_Zall.to_csv(self.outputZfile,sep=' ')
        betasToZero = list()
        ########################################################
        # Figure out from the reconstruction list which betas to zero 
        # Perform projection analysis. First standardize the data
        indexing = df_X.index.values
        headers = df_X.columns.values
        if (self.reconstructData):
            betasToZero = self.mapZtoBetas(self.inReconstructList,df_Z)
            print('Beta indexing for zeros '+str(betasToZero))
        # Scale features not cells
        if (self.scale and self.origScaling):
            sklearn.preprocessing.scale(df_X, axis=0, with_mean=True, with_std=True, copy=False)
        if (self.scale and not self.origScaling):
           print('Guess scaling: MinMax')
           scaler = MinMaxScaler(feature_range=(0.0,1.0)) # Might also want 0.1,0.9 instead
           df_X = pd.DataFrame(scaler.fit_transform(df_X))
        X = df_X.values
        Z = df_Z.values
        simdat = denseClassCovariateProjectionReductionReaddata(K,metadata)
        simdat.setXandYdata(X,Z)
        #NOTE X was be floating point ofr double prec
        if (self.reconstructData):
            simdat.specifyBetasToZero(betasToZero)
            VT,S,Szero = simdat.computeUandS_ZscoredEigenvaluesReturnReconstructed( betasToZero )
            print('Project using a Zscore-eigenvalue rank regression: Keep a reconstructed matrix')
        if (not self.reconstructData):
            if (runtype =='ZSCORE'):
                VT,S = simdat.computeUandS_ZscoredEigenvalues()
                print('Project using a Zscore-eigenvalue rank regression')
            else:
                VT,S = simdat.computeUandS()
                print('Project using eigenvalue rank')
        #
        self.U,self.d,self.VT,self.effectCorrection = simdat.returnSVDdata()
        #
        Xtilde = np.matmul(S,VT)
        df_Xtilde = pd.DataFrame(Xtilde)
        # Always scale the output result
        if (not self.origScaling):
            print('Guess scaling: MinMax')
            scaler = MinMaxScaler(feature_range=(0.0,1.0)) # Might also want 0.1,0.9 instead
            df_Xtilde = pd.DataFrame(scaler.fit_transform(df_Xtilde))
        else:
            sklearn.preprocessing.scale(df_Xtilde, axis=0, with_mean=True, with_std=True, copy=False)
        df_Xtilde['CELL']=indexing
        df_Xtilde.set_index('CELL',inplace=True)
        df_Xtilde.columns = headers
        print('write out new X with new metadata file')
        df_Xtilde.to_csv(self.outputXfile,sep=' ')
        self.df_Xtilde = df_Xtilde
        if (self.reconstructData):
            XtildeZero = np.matmul(Szero,VT)
            df_XtildeZero = pd.DataFrame(XtildeZero)
            if (not self.origScaling):
                print('Guess scaling: MinMax XtildeZero')
                scaler = MinMaxScaler(feature_range=(0.0,1.0)) # Might also want 0.1,0.9 instead
                df_XtildeZero = pd.DataFrame(scaler.fit_transform(df_XtildeZero))
            else:
                sklearn.preprocessing.scale(df_XtildeZero, axis=0, with_mean=True, with_std=True, copy=False)
            df_XtildeZero['CELL']=indexing
            df_XtildeZero.set_index('CELL',inplace=True)
            df_XtildeZero.columns = headers
            print('write out new reconstructed (some betas zeroed) X with new metadata file')
            df_XtildeZero.to_csv(self.outputXReconstructfile,sep=' ')
            self.df_XtildeZero = df_XtildeZero
        print('procedure completed')



#
# Construct the data sets given the sizes n,p,k,and w
#
import time as tm
import pandas as pd
import numpy as np
from broonie.project_group import projectGroupDataClass

#############################################################################
# Define functions for convergence testing
# X matrices include all effects: So we want to remove them all (again)

print('Begin imputation')

# Specify subspace for group effect reduction
k = 60

inputfilename='./DATA/initialGuess.tsv'

###################################################### group effect
# Straight up projection using the inlist and default effects
inlist='MOUSE_DATE_SEX_TREATMENT'
outtildemeta='testingOneXtilde_'
print('Run projection')
x = projectGroupDataClass(k,inputfilename,inlist,outtildemeta)
x.reportParameters()
x.runProjection()
xtilde = x.fetchXtilde() # Keep this for testing convergence
xfilename = x.getXtildeFilename()

######################################################
# Attempt to reconstruct TREATMENT projection using the inlist and default effects
inlist='MOUSE_DATE_SEX_TREATMENT'
outtildemeta='testingTwoXtilde_'
print('Run projection')
x = projectGroupDataClass(k,inputfilename,inlist,outtildemeta)
x.reconstructList('TREATMENT')
x.reportParameters()
x.runProjection()
xtilde = x.fetchXtilde() # Keep this for testing convergence
reconstructedXtilde = fetchReconstructedXtilde()
xfilename = x.getXtildeFilename()
reconstructedxfilename = getReconstructedXtilde()
print('Xtilde '+xfilename)
print('Reconstructed Xtilde '+reconstructedxfilename)

######################################################
# Use the filebased techniques for specifying possible effect names
# batchfile must contain ALL possible effects in the data set
# If batchfile list is incomplete errors from the SVD method will occur:
####
####AttributeError: 'str' object has no attribute 'conjugate'
####>>> xtilde = x.fetchXtilde() # Keep this for testing convergence
####>>> xfilename = x.getXtildeFilename()
####AttributeError: 'str' object has no attribute 'conjugate'
####>>> xtilde = x.fetchXtilde() # Keep this for testing convergence
####>>> xfilename = x.getXtildeFilename()

inlist='MOUSE_DATE_SEX_TREATMENT'
outtildemeta='testingThreeXtilde_'
batchfilename='listOfBatchEffects.txt'
x = projectGroupDataClass(k,inputfilename,inlist,outtildemeta,batchfilename)
x.reportParameters()
x.runProjection()
xtilde = x.fetchXtilde() # Keep this for testing convergence
xfilename = x.getXtildeFilename()


######################################################
# Use the filebased techniques for specifying possible effect names
# And build a reconstructed matrix

inlist='MOUSE_DATE_SEX_TREATMENT'
outtildemeta='testingFourXtilde_'
batchfilename='listOfBatchEffects.txt'
x = projectGroupDataClass(k,inputfilename,inlist,outtildemeta,batchfilename)
x.reconstructList('TREATMENT')
x.reportParameters()
x.runProjection()
xtilde = x.fetchXtilde() # Keep this for testing convergence
xfilename = x.getXtildeFilename()
reconstructedxfilename = getReconstructedXtilde()
print('Xtilde '+xfilename)
print('Reconstructed Xtilde '+reconstructedxfilename)


print('projection completed')





# Construct the data sets given the sizes n,p,k,and w
#
import time as tm
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from broonie.project_group import projectGroupDataClass
from broonie.generate_neighbor_matrix import generateNeighborMatrixClass
from broonie.regress import regressDataClass
from broonie.construct_initial_guess import constructInitialGuess
from sklearn.preprocessing import MinMaxScaler
import sys
from sklearn.externals.joblib import Parallel, delayed

#############################################################################
# Define functions for convergence testing
# X matrices include all effects: So we want to remove them all (again)

maxDiffElemsThreshold = 0.1
maxFrobeniusThreshold = 0.02
maxIterations = 12

##iterationMode = 'raw'
##iterationMode = 'incremental'

def dropAllEffects(df_X):
    currentPossibleColumnEffects = ['TREATMENT','MOUSE','SEX','DATE','NODEID']
    allColumns = df_X.columns.values
    foundColumns = list(set(allColumns) & set(currentPossibleColumnEffects))
    print('Number of check convergence: group effect columns in data is '+str(foundColumns))
    print('They are: '+str(foundColumns))
    return(df_X.drop(foundColumns,axis=1,inplace=False))

def checkConvergence(df_inX, df_inXold):
    """It has happened that the two input matrices have different row orders
    So we must address that here
    We do not anticipate checking Xtilde matrices. Rather R or S. Thus we must
    account for the possible covariates
    """
    print('in 1 '+str(df_inX.shape))
    print('in 2 '+str(df_inXold.shape))
    index1 = df_inX.index.values
    index2 = df_inXold.index.values
    if (index1.shape[0] != index2.shape[0]):
        print('different numbers of cells in the checkConvrgence: Abort')
        print(str(df_inX.shape))
        print(str(df_inXold.shape))
        sys.exit()
    combined = list(set(index1).intersection(index2))

    # first do they contain the same elements?

    converged = False
    df_X = dropAllEffects(df_inX.loc[combined]) 
    df_Xold = dropAllEffects(df_inXold.loc[combined])
    maxDiffElems =  (df_X-df_Xold).values.max()
    print('Max element '+str(maxDiffElems))
    if (maxDiffElems <= maxDiffElemsThreshold):
        print('X coverged on a maximum per-element threshold '+str(maxDiffElems))
        converged=True
    print('Max diff elem '+str(maxDiffElems))
    # Compute the frobenious norm  based convergence criteria
    print('try froben norm ')
    frac = np.linalg.norm(df_X-df_Xold) / np.linalg.norm(df_X)
    print('Convergence frac is '+str(frac))
    if (frac <= maxFrobeniusThreshold):
        print('X coverged on a frobenius threshold '+str(frac))
        converged=True
    print('Frobenious norm '+str(frac))
    return (converged)

def filterAmendedHeaders(allheaders):
    """ Remove any entry containing "encoded" or "ENC"
    These are hard-coded terms in the PROJECT methods. Beware
    """
    rawindices = [i for i in allheaders if not ('encoded' in i or 'ENC' in i)]
    return (rawindices)


def processMetaData(df_Z):
    allheaders = df_Z.columns.values 
    origheaders = filterAmendedHeaders(allheaders)
    return (allheaders,origheaders)

print('Begin imputation')

# Number of parallel jobs for the cluster code
# Presumably threadbased. So don.t go too big. 8 or 16 at the most
# at njobs=1 the jobs takes at ~15,000 secs and at njobs=16 total time is 1,903 sec.
njobs = 16 
# Specify subspace for group effect reduction
k = 60
# Specify number of NNs for regression (actually we use knn+1 becasue we include the target cell
knn = 10

# Initial Read data (R naught)

# RAW DATA
rawfilename='/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/PCAProjectionImputationPipeline/MouseTumor_allMice_allGenes_Rawdata/mouse_raw_data_allcells_Notscaled.tsv'

# Group effects to exclude at every iteration
inlist='MOUSE_DATE_SEX_TREATMENT'

# Fetch an initial guess. In this example we simply take the RAW data and also Read Ben's associzated "regressed" data.
# Bens data is the seurat processed scaling of the Raw data. Moving forward, we may want to process initial guesses differently

outfilename = 'initialGuess.tsv'
initialGuess = constructInitialGuess(rawfilename,outfilename,batchfile=False,origScaling=False,logScaledResults=False)
initialGuess.setOrigScaling(False)
initialGuess.reportScaling()
#initialGuess.runInitialGuess()
initialGuess.fetchInitialGuess() # Take Seurat imputed data rescale to MinMax (0,1) and proceed
initialGuessFilename = initialGuess.getOutputFilename() # return Ben's seurat regressed data set as initial guess
print('output bounds filename is '+initialGuess.getBoundsFilename())
scaleBounds = initialGuess.fetchScaleBounds()


# Initialize the iterative procedure 
inputfilename = initialGuessFilename

istart = 0 
converged=False

# Here we only do the projection once
outtildemeta='outputXtilde_New'
print('Run New projection')
x = projectGroupDataClass(k,inputfilename,inlist,outtildemeta,scale=True,origScaling=False) # regress studentizes already
x.reportParameters()
x.runProjection()
xtilde = x.fetchXtilde() # Keep this for testing convergence

xfilename = x.getXtildeFilename()

# IN this examp,e we wil not change the input file
inputregressionfilename = xfilename

inputXfilename = xfilename

#####################################################

Sold = None
Scurrent = xtilde 

while istart < maxIterations:
    #
    # if scaling by cells and features neede then do it here> Not required for initialGuess data
    print('Start the next new iteration'+str(istart))
###################################################### check convergence
# Note rows might be differently ordered
# Need to double check the test
# Check convergece of the R matrix ( One could also use the S matrix)
    if (Sold is not None):
        converged = checkConvergence(Scurrent, Sold)
    print('Convergence testing status is '+str(converged))
    if (converged):
        print('S has converged')
        break
##################################### Second. reclustering
# inputfilename='60_MOUSEDATESEXTREATMENT_Xtilde_ZscoreOrder.tsv'
# No group effects are included in this file

    outputIndices = 'dfIndices-'+str(istart)+'.tsv'
    outputDistances = 'dfDistances-'+str(istart)+'.tsv'
    print('Run clustering ' + inputXfilename)
    y = generateNeighborMatrixClass(knn,inputXfilename,outputIndices,outputDistances,njobs)
    y.reportParameters()
    y.runProcess()
    y.writeClusterInformation() # Write out the K and distance matrices
    yindexfilename = y.getIndexFilename()
    ydistancefilename = y.getDistanceFilename()
    kindices = y.reportIndices()
#################################### Third run the regression
    print('Run regressions')
    inputrawfilename = rawfilename
    zindexfilename = yindexfilename
    zdistancefilename = ydistancefilename
    outputRfilename ='newRoutput-'+str(istart)+'.tsv'
    outputSfilename ='newSoutput-'+str(istart)+'.tsv'
    Sold = Scurrent
    print(inputrawfilename)
    print(yindexfilename)
    print(ydistancefilename)
    #
    #use the oldS instead of raw R inputrawfilename
    #Inputfilename SHOULD not contain negative terms
    z = regressDataClass(knn,inputregressionfilename,yindexfilename,ydistancefilename,outputRfilename,outputSfilename,origScaling=False,logScaledResults=False)
    z.reportParameters()
    z.reportScaling()
    z.runProcess()
    z.writeNewR()
    z.writeNewS()
    #inputrawfilename = z.getOutputRfilename() # This still contains effects like the rawR
    #############################################################
    #
    xtildeprevious = xtilde # Hold for next convergence test
    # This approach results in osscilation of the Max elements
    Scurrent = z.getSmatrix()
    #Rcurrent = z.getRmatrix()

    #inputRmatrix = z.getRawRmatrix()

    ##inputregressionfilename = z.getOutputSfilename() # For regressing
    #inputregressionfilename = z.getOutputRfilename() # As another option
    inputXfilename = z.getOutputSfilename()# For clustering

    #if (iterationMode == 'raw'):
    #    inputregressionfilename = rawfilename
    #else:
    #    print('Iteration will impute always versus RAW R')
    #    inputregressionfilename = z.getOutputRfilename()
    #
    istart += 1

# If not converged proceed with gernating the final R. We could consider using this
# as an initial guess for the next run

if (converged):
    print('#################################  Iteration has converged')

#####################################
# Construct the final unscaled matrix with (linear) effects reintroiduced.
currentXtilde = z.getSmatrix()
cellIndex = currentXtilde.index.values
features = currentXtilde.columns.values

#####################################
# Get effects metadata
# NOTE we read the Z matric that contains BOTH the original effects data AND the filtered/encoded data
df_Zall = x.fetchZ()
allheaders,origheaders = processMetaData(df_Zall)
actualHeaders =  currentXtilde.columns.values # May not contain ANY effect terms but will have lots of features
intersectList = list(set(allheaders) & set(actualHeaders))
currentXtildeDropEffects = currentXtilde.drop(intersectList,axis=1) # we would expect no effects to be found

##################################### Fold Effects back into solution
# Must fuss around with including or not the covariates
# Do we keep df_new_S or df_new_R?
# S not scaled anymore since we now add effects back in
U,d,VT,effectCorrection = x.returnSVDdata()
df_new_S_temp = currentXtildeDropEffects + np.matmul(effectCorrection,VT)

# Rescale after including the effects
rescale = MinMaxScaler(feature_range=(0.0,1.0)) # Might also want 0.1,0.9 instead
df_new_S = pd.DataFrame(rescale.fit_transform(df_new_S_temp))
df_new_S.index = currentXtildeDropEffects.index
df_new_S.columns = currentXtildeDropEffects.columns
df_orig_Z = df_Zall[origheaders]
 
##################################### Inverse scale the solution
# Xpreminmax = y(max-min)+min
# THIS WILL NOT WORK for centerMean : I.e., origScaling=True
df_scaleBounds = initialGuess.fetchScaleBounds() 

# Take the effects+S terms that are scaled and unscale them
df_new_temp_unscaled = pd.DataFrame()
for i in df_scaleBounds.index:
    max = df_scaleBounds.loc[i]['MAX']
    min = df_scaleBounds.loc[i]['MIN']
    df_new_temp_unscaled[i] = [(j*(max-min)+min) for j in df_new_S[i]]

# Add original effect columns back in and we are finished
df_new_temp_unscaled.index = df_new_S.index
finalR_scaled = pd.merge(df_orig_Z,df_new_S,left_index=True,right_index=True)
finalR = pd.merge(df_orig_Z,df_new_temp_unscaled,left_index=True,right_index=True)

outputRimputed = 'Imputed_R_matrix_descaled_withEffectColumns.tsv'
finalR.to_csv(outputRimputed,sep=' ')
outputRimputedScaled = 'Imputed_R_matrix_scaled_withEffectColumns.tsv'
finalR_scaled.to_csv(outputRimputedScaled,sep=' ')

#
if (converged):
    print('Iterations has converged')
print('Exit the loop with number of iteration = '+str(istart))
print('imputation completed')
print('Final imputted descaled file is '+ outputRimputed) 
print('Final imputted scaled file is '+ outputRimputedScaled)




#################################################################
##
## Some elementary steps of the method by
## Aliverti, Lum, Johndrow, and Dunson.
## Attempt to build the simulation data set as instructed in 
## arXiv: 1810.08255v1 18,Oct,2018, pg 8
##
## Generate a simulated dataset.
## n: number of observations
## p: number of explanatory bariables
## k: number of dimensions 
## w: number of z terms that are "group effects" to be folded into the data (X) Z(nxw)
##
## X(nxp)
## Z(nxw)
## Y(nx1)
##
## X.Y data are centered and scaled to 1 on exit
##
## metadata: Base naming for output filename construction

##################################################################

import pandas as pd
import numpy as np
from numpy import zeros
from numpy import diag
import matplotlib.pyplot as plt
import sklearn.preprocessing
import random

class constructSimulatedDatasetCode(object):
    """Construct data set consisting of X,Y, and Z for testing the
    group projection method of Aliverti,Lum, Johndrow,and Dunson
    A Z is constructed of size nxw dimensions which is then folded into the X
    terms for subsequent removal testing

    Attributes
    ----------
    self.metadata = 'experimentOne'
    seld.metaparams = 'additional filename metadata :n-p-k-w
    self.numsubjects = n
    self.numvariables = p
    self.numcomponents = k
    self.numZterms = w
    """
    def __init__(self, metadata='testing', n=1000, p=100, k=10, w=1):
        self.metadata = metadata
        self.n = n
        self.p = p
        self.k = k
        self.w = w
        self.metaparams = str(self.n)+'_'+str(self.p)+'_'+str(self.k)+'_'+str(self.w)
        if (self.w > 1):
            print('Current code tested for SINGLE Z component' )
            #self.w = 1
        self.dfX = pd.DataFrame()
        self.dfY = pd.DataFrame()
        self.dfZ = pd.DataFrame()

    def outputSimulationParams(self):
        print('-------------------------------------------------------')
        print('Simulation parameters follows')
        print('Number of observations(n) = '+str(self.n))
        print('Number of variables(p) = '+str(self.p))
        print('Number of components(k) = '+str(self.k))
        print('Number of Z terms(w) = '+str(self.w))
        print('Filename meta data are '+self.metadata+'_'+self.metaparams)
        print('-------------------------------------------------------')

    def setn(self,inn):
        self.n = inn
    def setp(self,inp):
        self.p = inp
    def setk(self,ink):
        self.k = ink
    def setw (self,inw):
        self.w

##
## If ZI is one column then python subtracts ZI from EACH COLUMN OF S. IF ZI has same columns as S
## Pands does a column-wise subtrraction, otherwise they are nonconforming and the method fails
## So we must explicitly iterate tp subtract all W columns from S
    def buildXandYmatrix(self,S,W,Z,beta):
        """ Each row is drawn from a p-variate standard normal dist with 
        mean vector u_i = (s_i-2z_i1)W,i=1,n
        W are going to build the vector (MU) for all N simultaniously
        How do we handle the covariance part?
        """
        n = S.shape[0]
        p = W.shape[1]
        k = W.shape[0]
        w = Z.shape[1]
        SmZI = S
        if (Z.shape[1] != 1):
            print('Z should only have a single column for now '+str(Z.shape[1]))
        if (w > 1):
            for i in range(0,w):
                SmZI = SmZI - 2.0*Z[:,i].reshape(n,1)
        else:
            SmZI = SmZI - 2.0*Z
        mu_matrix = np.matmul(SmZI,W)
        mu_matrix_y = np.matmul(SmZI,beta)
        outputX = np.zeros((n,p),dtype=float)
        outputY = np.zeros((n,1),dtype=float)
        # Second part. Construct p-variate normal distributed row with mean vector meanrow
        icov = np.identity(p)
        for irow in range(0,n):
            meanrow = mu_matrix[irow]
            outputX[irow]=np.random.multivariate_normal(meanrow,icov,1)
            meanYrow = mu_matrix_y[irow]
            outputY[irow]=np.random.normal(loc = meanYrow,scale = 1)
        return outputX,outputY,mu_matrix,mu_matrix_y

    def constructDataObjects(self):
        """Construct the X,Y terms based on W,S,beta
        """
        n = self.n
        p = self.p
        k = self.k
        w = self.w

        # Build S(n*k) and W(k*p): independant normal sampled entries
        S = np.random.normal(0,1,size = (n,k))
        W = np.random.normal(0,1,size = (k,p))

        # Build indicator variable to remove
        Z = np.random.binomial(1,0.5,size=(n,w)) # Get me n samples of [0,1]
        beta = np.random.uniform(-5.0, 5.0,size=(k*1))
        beta = beta.reshape(k,1)

        # Build the X,Y matrix 
        Xprescaled,Yprescaled,mu_matrix,mu_matrix_y = self.buildXandYmatrix(S,W,Z,beta)

        # Center and Scale the X,Y matrices
        print('scale center/stddev the X data')

        X = sklearn.preprocessing.scale(Xprescaled,axis=0, with_mean=True, with_std=True, copy=True)
        Y = sklearn.preprocessing.scale(Yprescaled,axis=0, with_mean=True, with_std=True, copy=True)
        self.dfX = pd.DataFrame(X)
        self.dfY = pd.DataFrame(Y)
        self.dfZ = pd.DataFrame(Z)
        return

        #def checkHistogram(name,column):
        #    """For the name = dfX,dfY,dfZ
        #    choose the column (0,..) for display as a histrogram
        #    """
        #    if (name != 'dfX' and name != 'dfY' and name != 'dfZ'):
        #        print('Bad name to histrogram: Use only dfX,dfY,or dfZ')
        #        return
        #    if (column >= name.shape[1]):
        #        print('Column specification to large: must be less than '+str(name.shape[1]-1))
        #        return
        #    plt.plot(name[,column])
        #    plt.show()
        #    return

    def writeObjectsToDisk(self):
        """Write the new X,Y,Z terms to disk. X,Y are scaled
        """
        meta_params = self.metaparams
        metadata = self.metadata
        xfilename = metadata+'_'+meta_params+'Xdata.tsv'
        yfilename = metadata+'_'+meta_params+'Ydata.tsv'
        zfilename = metadata+'_'+meta_params+'Zdata.tsv'
        print('Writing simulated datasets')
        print('X file = '+xfilename)
        print('Y file = '+yfilename)
        print('Z file = '+zfilename)
        self.dfX.to_csv(xfilename,sep=' ')
        self.dfY.to_csv(yfilename,sep=' ')
        self.dfZ.to_csv(zfilename,sep=' ')
        print('finished with writing simulated datasets')

    def writeCombinedObjectToDisk(self):
        """Write the new X,and Z terms as a singlew entity to disk. X are scaled
        """
        combinedR = pd.merge(self.dfZ,self.dfX,left_index=True,right_index=True)
        combfilename = metadata+'_'+meta_params+'XandZdata.tsv'
        print('Writing simulated datasets')
        print('Writing simulated datasets')
        print('X file = '+xfilename)
        combinedR.to_csv(combfilename,sep=' ')







# Construct the data sets given the sizes n,p,k,and w
#
import time as tm
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from projectGroupDataClass import projectGroupDataClass
from generateNeighborMatrixClass import generateNeighborMatrixClass
from regressDataClass import regressDataClass
from constructInitialGuess import constructInitialGuess
import sys
from sklearn.externals.joblib import Parallel, delayed

#############################################################################
# Define functions for convergence testing
# X matrices include all effects: So we want to remove them all (again)

maxDiffElemsThreshold = 0.1
maxFrobeniusThreshold = 0.01
maxIterations = 12

##iterationMode = 'raw'
##iterationMode = 'incremental'

def dropAllEffects(df_X):
    currentPossibleColumnEffects = ['TREATMENT','MOUSE','SEX','DATE','NODEID']
    allColumns = df_X.columns.values
    foundColumns = list(set(allColumns) & set(currentPossibleColumnEffects))
    print('Number of check convergence: group effect columns in data is '+str(foundColumns))
    print('They are: '+str(foundColumns))
    return(df_X.drop(foundColumns,axis=1,inplace=False))

def checkConvergence(df_inX, df_inXold):
    """It has happened that the two input matrices have different row orders
    So we must address that here
    We do not anticipate checking Xtilde matrices. Rather R or S. Thus we must
    account for the possible covariates
    """
    print('in 1 '+str(df_inX.shape))
    print('in 2 '+str(df_inXold.shape))
    index1 = df_inX.index.values
    index2 = df_inXold.index.values
    if (index1.shape[0] != index2.shape[0]):
        print('different numbers of cells in the checkConvrgence: Abort')
        print(str(df_inX.shape))
        print(str(df_inXold.shape))
        sys.exit()
    combined = list(set(index1).intersection(index2))

    # first do they contain the same elements?

    converged = False
    df_X = dropAllEffects(df_inX.loc[combined]) 
    df_Xold = dropAllEffects(df_inXold.loc[combined])
    maxDiffElems =  (df_X-df_Xold).values.max()
    print('Max element '+str(maxDiffElems))
    if (maxDiffElems <= maxDiffElemsThreshold):
        print('X coverged on a maximum per-element threshold '+str(maxDiffElems))
        converged=True
    print('Max diff elem '+str(maxDiffElems))
    # Compute the frobenious norm  based convergence criteria
    print('try froben norm ')
    frac = np.linalg.norm(df_X-df_Xold) / np.linalg.norm(df_Xold)
    print('Convergence frac is '+str(frac))
    if (frac <= maxFrobeniusThreshold):
        print('X coverged on a frobenius threshold '+str(frac))
        converged=True
    print('Frobenious norm '+str(frac))
    return (converged)

def filterAmendedHeaders(allheaders):
    """ Remove any entry containing "encoded" or "ENC"
    These are hard-coded terms in the PROJECT methods. Beware
    """
    rawindices = [i for i in allheaders if not ('encoded' in i or 'ENC' in i)]
    return (rawindices)


def processMetaData(df_Z):
    allheaders = df_Z.columns.values 
    origheaders = filterAmendedHeaders(allheaders)
    return (allheaders,origheaders)

print('Begin imputation')

# Number of parallel jobs for the cluster code
# Presumably threadbased. So don.t go too big. 8 or 16 at the most
# at njobs=1 the jobs takes at ~15,000 secs and at njobs=16 total time is 1,903 sec.
njobs = 16 
# Specify subspace for group effect reduction
k = 60
# Specify number of NNs for regression (actually we use knn+1 becasue we include the target cell
knn = 10

# Initial Read data (R naught)

# RAW DATA
rawfilename='/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/PCAProjectionImputationPipeline/MouseTumor_allMice_allGenes_Rawdata/mouse_raw_data_allcells_Notscaled.tsv'

# Group effects to exclude at every iteration
inlist='MOUSE_DATE_SEX_TREATMENT'

# Fetch an initial guess. In this example we simply take the RAW data and also Read Ben's associzated "regressed" data.
# Bens data is the seurat processed scaling of the Raw data. Moving forward, we may want to process initial guesses differently

outfilename = 'initialGuess.tsv'
initialGuess = constructInitialGuess(rawfilename,outfilename,batchfile=False,origScaling=False,logScaledResults=False)
initialGuess.setOrigScaling(False)
initialGuess.reportScaling()
initialGuess.runInitialGuess()
initialGuessFilename = initialGuess.getOutputFilename() # return Ben's seurat regressed data set as initial guess
print('output bounds filename is '+initialGuess.getBoundsFilename())
scaleBounds = initialGuess.fetchScaleBounds()


# Initialize the iterative procedure 
inputfilename = initialGuessFilename

istart = 0 
converged=False

# Here we only do the projection once
outtildemeta='outputXtilde_New'
print('Run New projection')
x = projectGroupDataClass(k,inputfilename,inlist,outtildemeta,scale=True,origScaling=False) # regress studentizes already
x.reportParameters()
x.runProjection()
xtilde = x.fetchXtilde() # Keep this for testing convergence
xfilename = x.getXtildeFilename()

inputregressionfilename = xfilename

#####################################################

Sold = None
Scurrent = xtilde 

inputXfilename = xfilename

while istart < maxIterations:
    #
    # if scaling by cells and features neede then do it here> Not required for initialGuess data
    print('Start the next new iteration'+str(istart))
###################################################### check convergence
# Note rows might be differently ordered
# Need to double check the test
# Check convergece of the R matrix ( One could also use the S matrix)
    if (Sold is not None):
        converged = checkConvergence(Scurrent, Sold)
    print('Convergence testing status is '+str(converged))
    if (converged):
        print('S has converged')
        break
##################################### Second. reclustering
# inputfilename='60_MOUSEDATESEXTREATMENT_Xtilde_ZscoreOrder.tsv'
# No group effects are included in this file
    outputIndices = 'dfIndices-'+str(istart)+'.tsv'
    outputDistances = 'dfDistances-'+str(istart)+'.tsv'
    print('Run clustering ' + inputXfilename)
    y = generateNeighborMatrixClass(knn,inputXfilename,outputIndices,outputDistances,njobs)
    y.reportParameters()
    y.runProcess()
    y.writeClusterInformation() # Write out the K and distance matrices
    yindexfilename = y.getIndexFilename()
    ydistancefilename = y.getDistanceFilename()
    kindices = y.reportIndices()
#################################### Third run the regression
    print('Run regressions')
    inputrawfilename = rawfilename
    zindexfilename = yindexfilename
    zdistancefilename = ydistancefilename
    outputRfilename ='newRoutput-'+str(istart)+'.tsv'
    outputSfilename ='newSoutput-'+str(istart)+'.tsv'
    Sold = Scurrent
    print(inputrawfilename)
    print(yindexfilename)
    print(ydistancefilename)
    #
    #use the oldS instead of raw R inputrawfilename
    #Inputfilename SHOULD not contain negative terms
    z = regressDataClass(knn,inputregressionfilename,yindexfilename,ydistancefilename,outputRfilename,outputSfilename,origScaling=False,logScaledResults=False)
    z.reportParameters()
    z.reportScaling()
    z.runProcess()
    z.writeNewR()
    z.writeNewS()
    #inputrawfilename = z.getOutputRfilename() # This still contains effects like the rawR
    #############################################################
    #
    xtildeprevious = xtilde # Hold for next convergence test
    # This approach results in osscilation of the Max elements
    Scurrent = z.getSmatrix()
    #Rcurrent = z.getRmatrix()

    #inputRmatrix = z.getRawRmatrix()
    inputregressionfilename = z.getOutputSfilename() # Pass back to the OG methods
    inputXfilename = z.getOutputSfilename()# For clustering
    #if (iterationMode == 'raw'):
    #    inputregressionfilename = rawfilename
    #else:
    #    print('Iteration will impute always versus RAW R')
    #    inputregressionfilename = z.getOutputRfilename()
    #
    istart += 1

# If not converged proceed with gernating the final R. We could consider using this
# as an initial guess for the next run

if (converged):
    print('#################################  Iteration has converged')

#####################################
# Construct the final unscaled matrix with (linear) effects reintroiduced.
currentXtilde = Scurrent
cellIndex = currentXtilde.index.values
features = currentXtilde.columns.values

#####################################
# Get effects metadata
# NOTE we read the Z matric that contains BOTH the original effects data AND the filtered/encoded data
df_Zall = x.fetchZ()
allheaders,origheaders = processMetaData(df_Zall)

actualHeaders =  currentXtilde.columns.values # May not contain ANY effect terms but will have lots of features
intersectList = list(set(allheaders) & set(actualHeaders))
currentXtildeDropEffects = currentXtilde.drop(intersectList,axis=1)   # Generally we would expect no effects to be found

##################################### Fold Effects back into solution
# Must fuss around with including or not the covariates
# Do we keep df_new_S or df_new_R?

U,d,VT,effectCorrection = x.returnSVDdata()
df_new_S_temp = currentXtildeDropEffects + np.matmul(effectCorrection,VT)
df_orig_Z = df_Z[origheaders]
 
##################################### Inverse scale the solution
# Xpreminmax = y(max-min)+min
# THIS WILL NOT WORK for centerMean : I.e., origScaling=True
df_scaleBounds = initialGuess.fetchScaleBounds() 

# TODO Recast this into a high performance approach
df_new_temp_unscaled = pd.DataFrame()
for i in df_scaleBounds.index:
    max = df_scaleBounds.loc[i]['MAX']
    min = df_scaleBounds.loc[i]['MIN']
    df_new_temp_unscaled[i] = [(j-min)/(max-min) for j in df_new_S_temp[i]]

# Add original effect columns back in and we are finished

df_new_temp_unscale.index = df_new_S_temp.index

finalR_scaled = pd.merge(df_orig_Z,df_new_S_temp,left_index=True,right_index=True)
finalR = pd.merge(df_orig_Z,df_new_S_temp_unscale,left_index=True,right_index=True)

outputRimputed = 'Imputed_R_matrix_descaled_withEffectColumns.tsv'
finalR.to_csv(outputRimputed,sep=' ')

outputRimputedScaled = 'Imputed_R_matrix_scaled_withEffectColumns.tsv'
finalR_scaled.to_csv(outputRimputedScaled,sep=' ')

#
if (converged):
    print('Iterations has converged')
print('Exit the loop with number of iteration = '+str(istart))
print('imputation completed')
print('Final imputted descaled file is '+ outputRimputed) 
print('Final imputted scaled file is '+ outputRimputedScaled)




##
## Read in the input file including covariates. Then scale or otherwise proces
## the data such as by using seurat methods. In this current case, however, we
## will simply grab the "regressed" data from Ben and read it. Then we will dump out
## some statistices on the per-cell and per-gene normalizations

###########################################################################

## Basic input file format
## CELL SEX DATE NODEID MOUSE 0610007N19Rik ........

import pandas as pd
import numpy as np
import sys,ast
import matplotlib.pyplot as plt
import sklearn.preprocessing
from sklearn import preprocessing
from sklearn.preprocessing import MinMaxScaler

# Bens input data filename 
# /projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/PCAProjectionImputationPipeline/MouseTumor_allMice_allGenes/mouse_allPhenotypes_regressed_data_allCells_Notscaled.tsv

##########################################################################

#def getKnownGroupEffects(fulldata):
#    """ Check the input file for KNOWN possible group effects
#    Exclude those that may not exist such as TREATMENT. Make no
#    assumptions about order or sorting. Return list in arbitrary 
#    order. Independent of list provided by the user
#    """
#    currentPossibleColumnEffects = ['TREATMENT','MOUSE','SEX','DATE','NODEID']
#    allColumns = fulldata.columns.values
#    # Grab only those fulldata columns that exist inb currentPossible
#    foundColumns = list(set(allColumns) & set(currentPossibleColumnEffects))
#    print('Number of groupeffect columns in data is '+str(foundColumns))
#    print('They are: '+str(foundColumns))
#    return (foundColumns)


def getKnownGroupEffects(fulldata,batchfilename):
    """ Check the input file for KNOWN possible group effects
    Exclude those that may not exist such as TREATMENT. Make no
    assumptions about order or sorting. Return list in arbitrary 
    order. Independent of list provided by the user
    """
    foundColumns = list()
    listOneHotEncoded = list()
    if (batchfilename == False):
        print('Using prechosen batch effect names')
        currentPossibleColumnEffects = ['TREATMENT','MOUSE','SEX','DATE','NODEID']
        allColumns = fulldata.columns.values
        listOneHotEncoded = ['MOUSE'] # By default for use with our data testing
        # Grab only those fulldata columns that exist inb currentPossible
        foundColumns = list(set(allColumns) & set(currentPossibleColumnEffects))
    else:
        foundColumnsCombined = list()
        with open(batchfilename) as f:
            tempColumns = f.read().splitlines()
            for effect in tempColumns:
                #foundColumnsCombined.append(effect.upper())
                foundColumnsCombined.append(effect)
            for ipair in foundColumnsCombined:
                #print('effect is '+ipair)
                name,typeEnc = ipair.strip().split(' ')
                foundColumns.append(name)
                if (typeEnc=='ONEHOT'):
                    listOneHotEncoded.append(name)
    print('Number of Possible batch effects found/specified is '+str(len(foundColumns)))
    print('Number of groupeffect columns in data is '+str(foundColumns))
    #print('They have been converted to UPPER CASE: '+str(foundColumns))
    return (foundColumns,listOneHotEncoded)

def scaleResults(df_R,batchfile,origScaling,logScaledResults,cellScale):
    """ For the impute R (which includes group effects) scale 
    data for read counts, and then studentize the features to N(0,1)
    (value+1)*mean-all-cell-read / call total read
    The post averaged cell totals is approx 2100 for a factor of 1
    """
    scaleBounds = pd.DataFrame()
    covs,oneHotTerms = getKnownGroupEffects(df_R,batchfile)
    print(str(df_R.columns))
    print(str(covs))

    df_R_X_i=df_R.drop(covs,axis=1,inplace=False)
    df_new_R = df_R[covs] # Hold back until the end
    df_R_X = df_R_X_i.astype(float) # Copy false or not doesn't update inplace
    #

    factor = 10000.0

    # First ave number of reads per cell (or do we want median values kirk think median?)
    total_reads = float(df_R_X.values.sum())
    num_reads = float(df_R_X.values.size)
    ave_total_reads = total_reads / num_reads
    # Alternative method use the median
    ## total_sum_c = pd.DataFrame(df_R.sum(axis=1))
    ## median_total_reads = total_sum_c.median()
    ##total_median_c = pd.DataFrame(df_R.median(axis=1)) # or do we want median values
    ##total_median_c.columns=['TOTALC']

    # Sometimes the data come in as ints
    print('convert data type to float: use a mean method:')
    ##total_median_c = pd.DataFrame(df_R_X.median(axis=1)) # or do we want median values
    total_mean_c = pd.DataFrame(df_R_X.mean(axis=1)) # or do we want median values
    ##total_median_c.columns=['TOTALC']
    total_mean_c.columns=['TOTALC']
    # Ave number of reads per cells divided by cell reads
    total_correction_factor = factor * ave_total_reads / total_mean_c['TOTALC']
    # total_correction_factor.set_index('CELL')
    if (cellScale):
        print('Initial Guess: Will also scale cel total counts')
    ## Now we must also dump cell scaling info to disk
    ## diab le scaling cells

    ###############################################
    # Dump total correction data for subsequent de-scaling
    df_correction = pd.DataFrame(total_correction_factor)
    df_correction.columns = ['TOTALCORRECTION']
    df_correction.index = df_R_X.index

    if (cellScale):
        print('Preprocess cell scaling')
        if (logScaledResults):
            print('c+1 log scaling of counts')
            df_scale_new = np.log2((df_R_X+1.0).mul(total_correction_factor,axis=0))
        else:
            df_scale_new = df_R_X.mul(total_correction_factor,axis=0) # Scale cells to have same total reads
    else:
        df_scale_new = df_R_X

    # Lastly scale the features 
    # Getting lost when axis=1 and axis=0. Setting zero here gets us COLUMN scaling
    # Compute Max and Min values then store to disk for inverse scaling of final result
    ################################################
    # Dump min/max or mean/std FEATURE attributes to disk for subsequent inverse scaling
    if (origScaling):
        scaleBounds = pd.merge(df_scale_new.mean().to_frame(),df_new_scale.std().to_frame(),left_index=True,right_index=True)
        scaleBounds.columns = ['MEAN','STD']
    else:
        scaleBounds = pd.merge(df_scale_new.min().to_frame(),df_scale_new.max().to_frame(),left_index=True,right_index=True)
        scaleBounds.columns = ['MIN','MAX']
    ################################################
    if (origScaling):
        df_new = pd.DataFrame(sklearn.preprocessing.scale(df_scale_new, with_mean=True, with_std=True, copy=True,axis=0).tolist())
    else:
        print('Guess scaling: MinMax')
        scaler = MinMaxScaler(feature_range=(0.0,1.0)) # Might also want 0.1,0.9 instead
        df_new = pd.DataFrame(scaler.fit_transform(df_scale_new))

    # Now scale the cells
    #
    df_new.index = df_R_X.index.values
    df_new.columns = df_R_X.columns.values
    #
    ## diab le scaling cells
    ##if (logScaledResults):
    ##    df_scale_new = np.log((df_new+1.0).mul(total_correction_factor,axis=0))
    ##else:
    ##    df_scale_new = df_new.mul(total_correction_factor,axis=0) # Scale cells to have same total reads
    #df_scale_new = df_new

    # Add effects back into the data object

    finalR = pd.merge(df_new_R,df_new,left_index=True,right_index=True)
    df_new_R = finalR
    print('Scale the guess function')
    print(str(df_new_R.shape))
    print('---------------')
    return(df_new_R,scaleBounds,df_correction)

###############################################################################
# Set some globals

# May want to skip the first OG after initial guess construction

performSubsequentOG = True
batchfile=False

# Start the class

class constructInitialGuess(object):

    def reportParameters(self):
        print('Infilename is '+self.inputfilename)
        print('Output metadata '+self.outputInitialGuessfilename)

    def __init__(self, rawinfilename, initialGuessoutputfilename,batchfile=False,origScaling=True,logScaledResults=False,cellScaledResults=False):
        self.inputfilename = rawinfilename 
        self.batchfile = batchfile
        self.outputInitialGuessfilename = initialGuessoutputfilename
        self.origScaling = origScaling
        self.logScaledResults = logScaledResults if logScaledResults==False else True
        self.scaleBounds = pd.DataFrame()
        self.cellScaledResults = cellScaledResults if cellScaledResults==False else True
        self.totalCorrection = pd.DataFrame()
        self.outputBoundsfilename = 'scaleBounds.tsv'
        self.outputTotalCorrection = 'totalCorrection.tsv'
        self.preregressedfilename = '/projects/sequence_analysis/vol1/prediction_work/CausalInference/CausalNetworking_forKirk/PCAProjectionImputationPipeline/MouseTumor_allMice_allGenes/mouse_allPhenotypes_regressed_data_allCells_Notscaled.tsv'
        self.init = pd.DataFrame()

    def setOrigScaling(self,inOrigScale):
        if (inOrigScale==True):
            self.origScaling = True
        else:
            self.origScaling = False

    def reportScaling(self):
        print('origScaling is '+str(self.origScaling))
        print('Log scaling ? '+str(self.logScaledResults))
        print('Cell scaling ? '+str(self.cellScaledResults))

    def fetchScaleBounds(self):
        """Returns a dataframe of either MIN/MAX or MEAN/STD of the prestandardized features
        """
        return (self.scaleBounds)

    def fetchTotalCorrection(self):
        return (self.totalCorrection)

    def fetchInitialGuessMatrix(self):
        return (self.init)

    def getOutputFilename(self):
        return self.outputInitialGuessfilename
    
    def getBoundsFilename(self):
        return self.outputBoundsfilename

    def getTotalCorrectionFilename(self):
        if (self.cellScaledResults):
            return self.outputTotalCorrection
        else:
            return 'empty'

    def fetchInitialGuess(self):
        """Simply return Ben's Seurat imputed data set
        But read the data, and rescale it to MinMax (0,1) and rename to initialGuess.tsv
        """
        inputList = self.inputfilename
        inputpreregressedList = self.preregressedfilename
        regressdata  = pd.read_table(inputpreregressedList,delim_whitespace=True, index_col=0,header=0,low_memory=False)
        df_R,self.scaleBounds,self.totalCorrection = scaleResults(regressdata,self.batchfile,self.origScaling,self.logScaledResults,self.cellScaledResults)
        df_R.to_csv(self.outputInitialGuessfilename,sep=' ')
        print('Select Bens Seurat data file: Writing scale bounds '+self.getBoundsFilename())
        self.scaleBounds.to_csv(self.getBoundsFilename(),sep=' ')
        if (self.cellScaledResults):
            self.totalCorrection.to_csv(self.getTotalCorrectionFilename(),sep=' ')
        self.init = df_R
        print('Completed initial FETCH guess')

    def runInitialGuess(self):
        """
        Take input Reads matrix and scale it to mean cel counts AND ~N(0,1). No logs applied
        """
        inputList = self.inputfilename
        inputpreregressedList = self.preregressedfilename
        rawdata = pd.read_table(inputList,delim_whitespace=True,index_col=0,header=0,low_memory=False)
        df_R,self.scaleBounds,self.totalCorrection = scaleResults(rawdata,self.batchfile,self.origScaling,self.logScaledResults,self.cellScaledResults)
        self.init = df_R
        print('Writing guess to filename '+self.outputInitialGuessfilename)
        df_R.to_csv(self.outputInitialGuessfilename,sep=' ') 
        print('Writing scale bounds '+self.getBoundsFilename())
        self.scaleBounds.to_csv(self.getBoundsFilename(),sep=' ')
        if (self.cellScaledResults):
            print('Writing totalCorrection '+self.getTotalCorrectionFilename())
            self.totalCorrection.to_csv(self.getTotalCorrectionFilename(),sep=' ')
        print('Completed initial COMPUTE guess')
